import importlib
import json
from typing import List

from matplotlib.colors import LinearSegmentedColormap


class SciPalette:
    def __init__(self):
        """
        Cateogorical color palette collection of popular-sci journals.
        """
        handle = importlib.resources.open_binary("gseapy.data", "palette.json")
        self._db = json.load(handle)
        handle.close()

    def __repr__(self):
        # Name that Color. http://chir.ag/projects/name-that-color/
        return str(f"{self.__class__.__name__} color palette collection")

    def name_color(self, hex=None):
        """naming  a color

        :param hex: hex code of color
        """
        return "go to: http://chir.ag/projects/name-that-color/"

    @staticmethod
    def create_colormap(
        colors: List[str] = ["#000080", "#ffffff", "#8b0000"],
        positions: List[float] = [0.0, 0.5, 1.0],
        name: str = "navyblue2darkred",
    ):
        """create colormap given 3 color and break points

        :param colors: a list of color names. Default: ["#000080", "#ffffff", "#8b0000"]
        :param position: position of each color in range [0,1]. Default: [0.0, 0.5, 1.0]
        :param name: name of the return cmap
        :return: matplotlib cmap object
        """
        # cmap = ListedColormap(["#000080", "#ffffff", "#8b0000"])
        # positions = [0.0, 0.5, 1.0]
        if name is None:
            name = "mycmap"
        if colors is None:
            colors = ["#000080", "#ffffff", "#8b0000"]  # navyblue, white, darkred
        if positions is None:
            return LinearSegmentedColormap.from_list(name, colors)
        return LinearSegmentedColormap.from_list(name, list(zip(positions, colors)))

    @property
    def npg(self) -> List[str]:
        """Discrete Color Palettes inspired by plots in Nature Reviews Cancer"""
        return list(self._db["npg"].values())

    @property
    def aaas(self) -> List[str]:
        """Color palette inspired by plots in Science from AAAS"""
        return list(self._db["aaas"].values())

    @property
    def nejm(self) -> List[str]:
        """Color palette inspired by plots in The New England Journal of Medicine"""
        return list(self._db["nejm"].values())

    @property
    def lancet(self) -> List[str]:
        """Color palette inspired by plots in Lancet Oncology"""
        return list(self._db["lancet"].values())

    @property
    def jama(self) -> List[str]:
        """Color palette inspired by plots in The Journal of the American Medical Association"""
        return list(self._db["jama"].values())

    @property
    def jco(self) -> List[str]:
        """Color palette inspired by plots in Journal of Clinical Oncology"""
        return list(self._db["jco"].values())

    @property
    def ucscgb(self) -> List[str]:
        """Color palette inspired by UCSC Genome Browser Chromosome Colors"""
        return list(self._db["ucscgb"].values())

    def d3js(self, category: str = "c20a") -> List[str]:
        """
        choose category from (c10, c20a, c20b, c20c)
        """
        if category in self._db["d3js"]:
            return list(self._db["d3js"][category].values())
        return []

    @property
    def igv(self) -> List[str]:
        """Color palette inspired by IGV"""
        return list(self._db["igv"].values())

    @property
    def igv_alternating(self) -> List[str]:
        """Color palette inspired by IGV"""
        return list(self._db["igv_alternating"].values())

    @property
    def locuszoom(self) -> List[str]:
        """Color palette inspired by LocusZoom"""
        return list(self._db["locuszoom"].values())

    def uchicago(self, category: str = "default") -> List[str]:
        """
        Color palette inspired by University of Chicago Color Palette

        choose category from (light, dark, default)
        """
        if category in self._db["uchicago"]:
            return list(self._db["uchicago"][category].values())
        return []

    def hallmark(self, category: str = "dark") -> List[str]:
        """
        choose category from (dark, light)
        """
        if category in self._db["hallmark"]:
            return list(self._db["hallmark"][category].values())
        return []

    @property
    def cosmic(self) -> List[str]:
        """Color palette inspired by COSMIC Hallmarks of Cancer"""
        return list(self._db["cosmic"].values())

    @property
    def simpsons(self) -> List[str]:
        """Color palette inspired by The Simpsons"""
        return list(self._db["simpsons"].values())

    @property
    def futurama(self) -> List[str]:
        """Color palette inspired by Futurama"""
        return list(self._db["futurama"].values())

    @property
    def rickandmorty(self) -> List[str]:
        """Color palette inspired by Rick and Morty"""
        return list(self._db["rickandmorty"].values())

    @property
    def startrek(self) -> List[str]:
        """Color palette inspired by Star Trek"""
        return list(self._db["startrek"].values())

    @property
    def tron(self) -> List[str]:
        """Color palette inspired by Tron Legacy"""
        return list(self._db["tron"].values())

    @property
    def gsea(self) -> List[str]:
        """Color palette inspired by heatmaps generated by GSEA GenePattern"""
        return list(self._db["gsea"].values())

    def material(self, category: str = "indigo") -> List[str]:
        """
        choose category from
        (red, pink, purple, deeppurple, blue, lightblue, indigo,
        cyan, teal, lime, green, yellow, amber, organge, deeporange,
        brown, gray, bluegray)
        """
        if category in self._db["material"]:
            return list(self._db["material"][category].values())
        return []

    @property
    def zeileis(self) -> List[str]:
        """
        https://graphicdesign.stackexchange.com/questions/3682/where-can-i-find-a-large-palette-set-of-contrasting-colors-for-coloring-many-d
        orig reference http://epub.wu.ac.at/1692/1/document.pdf
        """
        return list(self._db["zeileis"].values())

    @property
    def godsnot(self) -> List[str]:
        """
        take from http://godsnotwheregodsnot.blogspot.de/2012/09/color-distribution-methodology.html
        or
        http://godsnotwheregodsnot.blogspot.com/2013/11/kmeans-color-quantization-seeding.html
        """
        return list(self._db["godsnot"].values())

    @property
    def boynton(self) -> List[str]:
        return list(self._db["boynton"].values())

    @property
    def kelly(self) -> List[str]:
        """
        The set of 22 colours of maximum contrast proposed by Kenneth Kelly in the work:
        http://www.iscc-archive.org/pdf/PC54_1724_001.pdf
        """
        return list(self._db["kelly"].values())

    @property
    def watlington(self) -> List[str]:
        return list(self._db["watlington"].values())

    @property
    def glasbey(self) -> List[str]:
        return list(self._db["glasbey"].values())



# -*- coding: utf-8 -*-

import logging
from math import ceil
from typing import Dict, Iterable, List

import numpy as np
from joblib import Parallel, delayed


def enrichment_score(
    gene_list: Iterable[str],
    correl_vector: Iterable[float],
    gene_set: Dict[str, List[str]],
    weight: float = 1.0,
    nperm: int = 1000,
    seed: int = 123,
    single: bool = False,
    scale: bool = False,
):
    """This is the most important function of GSEApy. It has the same algorithm with GSEA and ssGSEA.

    :param gene_list:       The ordered gene list gene_name_list, rank_metric.index.values
    :param gene_set:        gene_sets in gmt file, please use gsea_gmt_parser to get gene_set.
    :param weight:  It's the same with gsea's weighted_score method. Weighting by the correlation
                            is a very reasonable choice that allows significant gene sets with less than perfect coherence.
                            options: 0(classic),1,1.5,2. default:1. if one is interested in penalizing sets for lack of
                            coherence or to discover sets with any type of nonrandom distribution of tags, a value p < 1
                            might be appropriate. On the other hand, if one uses sets with large number of genes and only
                            a small subset of those is expected to be coherent, then one could consider using p > 1.
                            Our recommendation is to use p = 1 and use other settings only if you are very experienced
                            with the method and its behavior.

    :param correl_vector:   A vector with the correlations (e.g. signal to noise scores) corresponding to the genes in
                            the gene list. Or rankings, rank_metric.values
    :param nperm:           Only use this parameter when computing esnull for statistical testing. Set the esnull value
                            equal to the permutation number.
    :param seed:            Random state for initializing gene list shuffling. Default: seed=None

    :return:

    ES: Enrichment score (real number between -1 and +1)

    ESNULL: Enrichment score calculated from random permutations.

    Hits_Indices: Index of a gene in gene_list, if gene is included in gene_set.

    RES: Numerical vector containing the running enrichment score for all locations in the gene list .

    """
    N = len(gene_list)
    # Test whether each element of a 1-D array is also present in a second array
    # It's more intuitive here than original enrichment_score source code.
    # use .astype to covert bool to integer
    tag_indicator = np.in1d(gene_list, gene_set, assume_unique=True).astype(
        int
    )  # notice that the sign is 0 (no tag) or 1 (tag)

    if weight == 0:
        correl_vector = np.repeat(1, N)
    else:
        correl_vector = np.abs(correl_vector) ** weight

    # get indices of tag_indicator
    hit_ind = np.flatnonzero(tag_indicator).tolist()
    # if used for compute esnull, set esnull equal to permutation number, e.g. 1000
    # else just compute enrichment scores
    # set axis to 1, because we have 2D array
    axis = 1
    tag_indicator = np.tile(tag_indicator, (nperm + 1, 1))
    correl_vector = np.tile(correl_vector, (nperm + 1, 1))
    # gene list permutation
    rs = np.random.RandomState(seed)
    for i in range(nperm):
        rs.shuffle(tag_indicator[i])
    # np.apply_along_axis(rs.shuffle, 1, tag_indicator)

    Nhint = tag_indicator.sum(axis=axis, keepdims=True)
    sum_correl_tag = np.sum(correl_vector * tag_indicator, axis=axis, keepdims=True)
    # compute ES score, the code below is identical to gsea enrichment_score method.
    no_tag_indicator = 1 - tag_indicator
    Nmiss = N - Nhint
    norm_tag = 1.0 / sum_correl_tag
    norm_no_tag = 1.0 / Nmiss

    RES = np.cumsum(
        tag_indicator * correl_vector * norm_tag - no_tag_indicator * norm_no_tag,
        axis=axis,
    )

    if scale:
        RES = RES / N
    if single:
        es_vec = RES.sum(axis=axis)
    else:
        max_ES, min_ES = RES.max(axis=axis), RES.min(axis=axis)
        es_vec = np.where(np.abs(max_ES) > np.abs(min_ES), max_ES, min_ES)
    # extract values
    es, esnull, RES = es_vec[-1], es_vec[:-1], RES[-1, :]

    return es, esnull, hit_ind, RES


def enrichment_score_tensor(
    gene_mat,
    cor_mat,
    gene_sets,
    weighted_score_type,
    nperm=1000,
    seed=None,
    single=False,
    scale=False,
):
    """Next generation algorithm of GSEA and ssGSEA. Works for 3d array

    :param gene_mat:        the ordered gene list(vector) with or without gene indices matrix.
    :param cor_mat:         correlation vector or matrix  (e.g. signal to noise scores)
                            corresponding to the genes in the gene list or matrix.
    :param dict gene_sets:  gmt file dict.
    :param float weighted_score_type:     weighting by the correlation.
                            options: 0(classic), 1, 1.5, 2. default:1 for GSEA and 0.25 for ssGSEA.
    :param int nperm:       permutation times.
    :param bool scale:      If True, normalize the scores by number of genes_mat.
    :param bool single:     If True, use ssGSEA algorithm, otherwise use GSEA.
    :param seed:              Random state for initialize gene list shuffling.
                            Default: seed=None
    :return: a tuple contains::

             | ES: Enrichment score (real number between -1 and +1), for ssGSEA, set scale eq to True.
             | ESNULL: Enrichment score calculated from random permutation.
             | Hits_Indices: Indices of genes if genes are included in gene_set.
             | RES: The running enrichment score for all locations in the gene list.

    """
    rs = np.random.RandomState(seed)
    # gene_mat -> 1d: prerank, ssSSEA or 2d: GSEA
    keys = sorted(gene_sets.keys())

    if weighted_score_type == 0:
        # don't bother doing calcuation, just set to 1
        cor_mat = np.ones(cor_mat.shape)
    elif weighted_score_type > 0:
        pass
    else:
        logging.error("Using negative values of weighted_score_type, not allowed")
        raise ValueError("weighted_score_type should be postive numerics")

    cor_mat = np.abs(cor_mat)
    if cor_mat.ndim == 1:
        # ssGSEA or Prerank
        # genestes->M, genes->N, perm-> axis=2
        N, M = len(gene_mat), len(keys)
        # generate gene hits matrix
        # for 1d ndarray of gene_mat, set assume_unique=True,
        # means the input arrays are both assumed to be unique,
        # which can speed up the calculation.
        tag_indicator = np.vstack(
            [np.in1d(gene_mat, gene_sets[key], assume_unique=True) for key in keys]
        )
        tag_indicator = tag_indicator.astype(int)
        # index of hits
        hit_ind = [np.flatnonzero(tag).tolist() for tag in tag_indicator]
        # generate permutated hits matrix
        perm_tag_tensor = np.repeat(tag_indicator, nperm + 1).reshape((M, N, nperm + 1))
        # shuffle matrix, last matrix is not shuffled when nperm > 0
        if nperm:
            np.apply_along_axis(
                lambda x: np.apply_along_axis(rs.shuffle, 0, x),
                1,
                perm_tag_tensor[:, :, :-1],
            )
        # missing hits
        no_tag_tensor = 1 - perm_tag_tensor
        # calculate numerator, denominator of each gene hits
        rank_alpha = (
            perm_tag_tensor * cor_mat[np.newaxis, :, np.newaxis]
        ) ** weighted_score_type

    elif cor_mat.ndim == 2:
        # GSEA
        # 2d ndarray, gene_mat and cor_mat are shuffled already
        # reshape matrix
        cor_mat = cor_mat.T
        # gene_mat is a tuple contains (gene_name, permuate_gene_name_indices)
        genes, genes_ind = gene_mat
        # genestes->M, genes->N, perm-> axis=2
        # don't use assume_unique=True in 2d array when use np.isin().
        # elements in gene_mat are not unique, or will cause unwanted results
        tag_indicator = np.vstack(
            [np.in1d(genes, gene_sets[key], assume_unique=True) for key in keys]
        )
        tag_indicator = tag_indicator.astype(int)
        perm_tag_tensor = np.stack(
            [tag.take(genes_ind).T for tag in tag_indicator], axis=0
        )
        # index of hits
        hit_ind = [np.flatnonzero(tag).tolist() for tag in perm_tag_tensor[:, :, -1]]
        # nohits
        no_tag_tensor = 1 - perm_tag_tensor
        # calculate numerator, denominator of each gene hits
        rank_alpha = (
            perm_tag_tensor * cor_mat[np.newaxis, :, :]
        ) ** weighted_score_type
    else:
        logging.error("Program die because of unsupported input")
        raise ValueError("Correlation vector or matrix (cor_mat) is not supported")

    # Nhint = tag_indicator.sum(1)
    # Nmiss =  N - Nhint
    axis = 1
    P_GW_denominator = np.sum(rank_alpha, axis=axis, keepdims=True)
    P_NG_denominator = np.sum(no_tag_tensor, axis=axis, keepdims=True)
    REStensor = np.cumsum(
        rank_alpha / P_GW_denominator - no_tag_tensor / P_NG_denominator, axis=axis
    )
    # ssGSEA: scale es by gene numbers ?
    # https://gist.github.com/gaoce/39e0907146c752c127728ad74e123b33
    if scale:
        REStensor = REStensor / len(gene_mat)
    if single:
        # ssGSEA
        esmatrix = REStensor.sum(axis=axis)
    else:
        # GSEA
        esmax, esmin = REStensor.max(axis=axis), REStensor.min(axis=axis)
        esmatrix = np.where(np.abs(esmax) > np.abs(esmin), esmax, esmin)

    es, esnull, RES = esmatrix[:, -1], esmatrix[:, :-1], REStensor[:, :, -1]

    return es, esnull, hit_ind, RES


def fast_ssgsea(tag_indicator, ranking):
    """
    tag_indicator: see `enrichment_score()`
    ranking: ranking values
    """
    n = len(ranking)  # number of genes
    k = tag_indicator.sum()  # number of overalped genes
    idxs = np.flatnonzero(tag_indicator)  # indices low to high
    step_cdf_in = np.sum(ranking[idxs] * (n - idxs)) / np.sum(ranking[idxs])
    step_cdf_out = (n * (n + 1) / 2 - np.sum(n - idxs)) / (n - k)
    es = step_cdf_in - step_cdf_out
    return es


def ranking_metric_tensor(
    exprs,
    method,
    permutation_num,
    pos,
    neg,
    classes,
    ascending,
    seed=None,
    skip_last=False,
):
    """Build shuffled ranking matrix when permutation_type eq to phenotype.
    Works for 3d array.

    :param exprs:   gene_expression DataFrame, gene_name indexed.
    :param str method:  calculate correlation or ranking. methods including:
                        1. 'signal_to_noise' (s2n) or 'abs_signal_to_noise' (abs_s2n).
                        2. 't_test'.
                        3. 'ratio_of_classes' (also referred to as fold change).
                        4. 'diff_of_classes'.
                        5. 'log2_ratio_of_classes'.
    :param int permuation_num: how many times of classes is being shuffled
    :param str pos: one of labels of phenotype's names.
    :param str neg: one of labels of phenotype's names.
    :param list classes:  a list of phenotype labels, to specify which column of
                          dataframe belongs to what class of phenotype.
    :param bool ascending:  bool. Sort ascending vs. descending.
    :param seed: random_state seed
    :param bool skip_last: (internal use only) whether to skip the permutation of the last rankings.

    :return:
             returns two 2d ndarray with shape (nperm, gene_num).

             | cor_mat_indices: the indices of sorted and permutated (exclude last row) ranking matrix.
             | cor_mat: sorted and permutated (exclude last row) ranking matrix.

    """
    rs = np.random.RandomState(seed)
    # S: samples, G: gene number
    G, S = exprs.shape
    # genes = exprs.index.values
    expr_mat = exprs.values.T
    perm_cor_tensor = np.tile(expr_mat, (permutation_num, 1, 1))
    if skip_last:
        # random shuffle on the first dim, the last matrix (expr_mat) is not shuffled
        for arr in perm_cor_tensor[:-1]:
            rs.shuffle(arr)
    else:
        for arr in perm_cor_tensor:
            rs.shuffle(arr)
    # metrics
    classes = np.array(classes)
    pos = classes == pos
    neg = classes == neg
    n_pos = np.sum(pos)
    n_neg = np.sum(neg)
    pos_cor_mean = perm_cor_tensor[:, pos, :].mean(axis=1)
    neg_cor_mean = perm_cor_tensor[:, neg, :].mean(axis=1)
    pos_cor_std = perm_cor_tensor[:, pos, :].std(axis=1, ddof=1)
    neg_cor_std = perm_cor_tensor[:, neg, :].std(axis=1, ddof=1)

    if method in ["signal_to_noise", "s2n"]:
        cor_mat = (pos_cor_mean - neg_cor_mean) / (pos_cor_std + neg_cor_std)
    elif method in ["abs_signal_to_noise", "abs_s2n"]:
        cor_mat = np.abs((pos_cor_mean - neg_cor_mean) / (pos_cor_std + neg_cor_std))
    elif method == "t_test":
        denom = np.sqrt((pos_cor_std**2) / n_pos + (neg_cor_std**2) / n_neg)
        cor_mat = (pos_cor_mean - neg_cor_mean) / denom
    elif method == "ratio_of_classes":
        cor_mat = pos_cor_mean / neg_cor_mean
    elif method == "diff_of_classes":
        cor_mat = pos_cor_mean - neg_cor_mean
    elif method == "log2_ratio_of_classes":
        cor_mat = np.log2(pos_cor_mean / neg_cor_mean)
    else:
        logging.error("Please provide correct method name!!!")
        raise LookupError("Input method: %s is not supported" % method)
    # return matix[nperm+1, perm_cors]
    cor_mat_ind = cor_mat.argsort()
    # ndarray: sort in place
    cor_mat.sort()
    # genes_mat = genes.take(cor_mat_ind)
    if ascending:
        return cor_mat_ind, cor_mat
    # descending order of ranking and genes
    # return genes_mat[:,::-1], cor_mat[:,::-1]
    return cor_mat_ind[:, ::-1], cor_mat[:, ::-1]


def gsea_compute_tensor(
    data,
    gmt,
    n,
    weighted_score_type,
    permutation_type,
    method,
    pheno_pos,
    pheno_neg,
    classes,
    ascending,
    processes=1,
    seed=None,
    single=False,
    scale=False,
):
    """compute enrichment scores and enrichment nulls.
    This function will split large array into smaller pieces to advoid memroy overflow.

     :param data: preprocessed expression dataframe or a pre-ranked file if prerank=True.
     :param dict gmt: all gene sets in .gmt file. need to call load_gmt() to get results.
     :param int n: permutation number. default: 1000.
     :param str method: ranking_metric method. see above.
     :param str pheno_pos: one of labels of phenotype's names.
     :param str pheno_neg: one of labels of phenotype's names.
     :param list classes: a list of phenotype labels, to specify which column of dataframe belongs to what category of phenotype.
     :param float weighted_score_type: default:1
     :param bool ascending: sorting order of rankings. Default: False.
     :param seed: random seed. Default: np.random.RandomState()
     :param bool scale: if true, scale es by gene number.

     :return: a tuple contains::

             | zipped results of es, nes, pval, fdr.
             | nested list of hit indices of input gene_list.
             | nested list of ranked enrichment score of each input gene_sets.
             | list of enriched terms

    """
    w = weighted_score_type
    subsets = sorted(gmt.keys())
    genes_mat, cor_mat = data.index.values, data.values
    base = 5 if data.shape[0] >= 5000 else 10
    ## phenotype permutation
    np.random.seed(seed)  # control the ranodm numbers

    if permutation_type == "phenotype":
        # shuffling classes and generate random correlation rankings
        logging.debug("Start to permutate classes..............................")

        if (n + 1) % base == 0:  # n+1: last permute is for orignial ES calculation
            num_bases = [base] * ((n + 1) // base)
            skip_last = [0] * (n // base) + [1]  # last is not permuted
        else:
            num_bases = [base] * ((n + 1) // base) + [(n + 1) % base]
            skip_last = [0] * ((n + 1) // base) + [(n + 1) % base]
        random_seeds = np.random.randint(np.iinfo(np.int32).max, size=len(num_bases))
        genes_ind = []
        cor_mat = []
        # split permutation array into smaller blocks to save memory
        temp_rnk = Parallel(n_jobs=processes)(
            delayed(ranking_metric_tensor)(
                data, method, b, pheno_pos, pheno_neg, classes, ascending, se, skip
            )
            for b, skip, se in zip(num_bases, skip_last, random_seeds)
        )

        for k, temp in enumerate(temp_rnk):
            gi, cor = temp
            genes_ind.append(gi)
            cor_mat.append(cor)
        genes_ind, cor_mat = np.vstack(genes_ind), np.vstack(cor_mat)
        # convert to tuple
        genes_mat = (data.index.values, genes_ind)

    logging.debug("Start to compute es and esnulls........................")
    # Prerank, ssGSEA, GSEA
    es = []
    RES = []
    hit_ind = []
    esnull = []
    temp_esnu = []

    # split gmt dataset, too
    block = ceil(len(subsets) / base)
    random_seeds = np.random.randint(np.iinfo(np.int32).max, size=block)
    # split large array into smaller blocks to avoid memory overflow
    i, m = 1, 0
    gmt_block = []
    while i <= block:
        # you have to reseed, or all your processes are sharing the same seed value
        rs = random_seeds[i - 1]
        gmtrim = {k: gmt.get(k) for k in subsets[m : base * i]}
        gmt_block.append(gmtrim)
        m = base * i
        i += 1
    ## if permutation_type == "phenotype": n = 0
    ## NOTE for GSEA: cor_mat is 2d array, it won't permute again when call enrichment_score_tensor
    temp_esnu = Parallel(n_jobs=processes)(
        delayed(enrichment_score_tensor)(
            genes_mat, cor_mat, gmtrim, w, n, rs, single, scale
        )
        for gmtrim, rs in zip(gmt_block, random_seeds)
    )

    # esn is a list, don't need to use append method.
    for si, temp in enumerate(temp_esnu):
        # e, enu, hit, rune = temp.get()
        e, enu, hit, rune = temp
        esnull.append(enu)
        es.append(e)
        RES.append(rune)
        hit_ind += hit
    # concate results
    es, esnull, RES = np.hstack(es), np.vstack(esnull), np.vstack(RES)

    return gsea_significance(es, esnull), hit_ind, RES, subsets


def gsea_compute(
    data,
    gmt,
    n,
    weighted_score_type,
    permutation_type,
    method,
    pheno_pos,
    pheno_neg,
    classes,
    ascending,
    processes=1,
    seed=None,
    single=False,
    scale=False,
):
    """compute enrichment scores and enrichment nulls.

    :param data: preprocessed expression dataframe or a pre-ranked file if prerank=True.
    :param dict gmt: all gene sets in .gmt file. need to call load_gmt() to get results.
    :param int n: permutation number. default: 1000.
    :param str method: ranking_metric method. see above.
    :param str pheno_pos: one of labels of phenotype's names.
    :param str pheno_neg: one of labels of phenotype's names.
    :param list classes: a list of phenotype labels, to specify which column of dataframe belongs to what category of phenotype.
    :param float weighted_score_type: default:1
    :param bool ascending: sorting order of rankings. Default: False.
    :param seed: random seed. Default: np.random.RandomState()
    :param bool scale: if true, scale es by gene number.

    :return: a tuple contains::

            | zipped results of es, nes, pval, fdr.
            | nested list of hit indices of input gene_list.
            | nested list of ranked enrichment score of each input gene_sets.
            | list of enriched terms

    """

    w = weighted_score_type
    subsets = sorted(gmt.keys())
    es = []
    RES = []
    hit_ind = []
    esnull = [[] for a in range(len(subsets))]
    np.random.seed(seed)  # control the ranodm numbers
    logging.debug("Start to compute enrichment scores......................")

    if permutation_type == "phenotype":
        logging.debug("Start to permutate classes..............................")
        # this version won't split large array into smaller ones
        genes_mat, cor_mat = ranking_metric_tensor(
            exprs=data,
            method=method,
            permutation_num=n + 1,
            pos=pheno_pos,
            neg=pheno_neg,
            classes=classes,
            ascending=ascending,
            seed=seed,
            skip_last=True,
        )

        # compute es, esnulls. hits, RES
        logging.debug("Start to compute enrichment nulls.......................")
        es, esnull, hit_ind, RES = enrichment_score_tensor(
            gene_mat=genes_mat,
            cor_mat=cor_mat,
            gene_sets=gmt,
            weighted_score_type=w,
            nperm=n,
            seed=seed,
            single=False,
            scale=False,
        )

    else:
        # Prerank, ssGSEA, GSEA with gene_set permutation
        gl, cor_vec = data.index.values, data.values
        logging.debug("Start to compute es and esnulls........................")
        ## this version don't split large array into smaller ones
        # es, esnull, hit_ind, RES = enrichment_score_tensor(gene_mat=gl,
        #                                                    cor_mat=cor_vec,
        #                                                    gene_sets=gmt,
        #                                                    weighted_score_type=w,
        #                                                    nperm=n, rs=rs
        #                                                    single=single, scale=scale)
        temp_esnu = []
        # you have to reseed, or all your processes are sharing the same seed value
        # np.random.seed(seed)
        random_seeds = np.random.randint(np.iinfo(np.int32).max, size=len(subsets))
        temp_esnu = Parallel(n_jobs=processes)(
            delayed(enrichment_score)(
                gl, cor_vec, gmt.get(subset), w, n, rs, single, scale
            )
            for subset, rs in zip(subsets, random_seeds)
        )
        # esn is a list, don't need to use append method.
        for si, temp in enumerate(temp_esnu):
            e, enu, hit, rune = temp
            esnull[si] = enu
            es.append(e)
            RES.append(rune)
            hit_ind.append(hit)

    return gsea_significance(es, esnull), hit_ind, RES, subsets


def normalize(es, esnull):
    """normalize the ES(S,pi) and the observed ES(S), separately rescaling
    the positive and negative scores by dividing the mean of the ES(S,pi).

    return: NES, NESnull
    """

    nEnrichmentScores = np.zeros(es.shape)
    nEnrichmentNulls = np.zeros(esnull.shape)
    # esnullmean = np.zeros(es.shape)
    # # calculate nESnulls
    # for i in range(esnull.shape[0]):
    #     # NES
    #     enrNull = esnull[i]
    #     if es[i] >= 0:
    #         mes = enrNull[enrNull >= 0].mean()
    #         nEnrichmentScores[i] = es[i] / mes
    #     else:
    #         mes = enrNull[enrNull < 0 ].mean()
    #         nEnrichmentScores[i] = - es[i] / mes
    #     esnullmean[i] = mes

    #     # NESnull
    #     for j in range(esnull.shape[1]):
    #         if esnull[i,j] >= 0:
    #             nEnrichmentNulls[i,j] = esnull[i,j] / esnullmean[i]
    #         else:
    #             nEnrichmentNulls[i,j] = - esnull[i,j] / esnullmean[i]

    esnull_pos = np.ma.MaskedArray(esnull, mask=(esnull < 0)).mean(axis=1)
    esnull_neg = np.ma.MaskedArray(esnull, mask=(esnull >= 0)).mean(axis=1)
    esnull_pos = np.array(esnull_pos)
    esnull_neg = np.array(esnull_neg)
    # NES
    nEnrichmentScores = np.where(es >= 0, es / esnull_pos, -es / esnull_neg)
    # NES_NULL
    nEnrichmentNulls = np.where(
        esnull >= 0,
        esnull / esnull_pos[:, np.newaxis],
        -esnull / esnull_neg[:, np.newaxis],
    )

    return nEnrichmentScores, nEnrichmentNulls


def gsea_pval(es, esnull):
    """Compute nominal p-value.

    From article (PNAS):
    estimate nominal p-value for S from esnull by using the positive
    or negative portion of the distribution corresponding to the sign
    of the observed ES(S).
    """

    # to speed up, using numpy function to compute pval in parallel.
    condlist = [es < 0, es >= 0]
    choicelist = [
        (esnull < es.reshape(len(es), 1)).sum(axis=1) / (esnull < 0).sum(axis=1),
        (esnull >= es.reshape(len(es), 1)).sum(axis=1) / (esnull >= 0).sum(axis=1),
    ]
    pvals = np.select(condlist, choicelist)

    return pvals


def gsea_fdr(nEnrichmentScores, nEnrichmentNulls):
    """Create a histogram of all NES(S,pi) over all S and pi.
       Use this null distribution to compute an FDR q value.

    :param nEnrichmentScores:  normalized ES
    :param nEnrichmentNulls:   normalized ESnulls
    :return: FDR
    """

    # FDR null distribution histogram
    # vals = reduce(lambda x,y: x+y, nEnrichmentNulls, [])
    # nvals = np.array(sorted(vals))
    # or
    nvals = np.sort(nEnrichmentNulls.flatten())
    nnes = np.sort(nEnrichmentScores)
    fdrs = []
    # FDR computation
    for i in range(len(nEnrichmentScores)):
        nes = nEnrichmentScores[i]
        # use the same pval method to calculate fdr
        if nes >= 0:
            allPos = int(len(nvals) - np.searchsorted(nvals, 0, side="left"))
            allHigherAndPos = int(len(nvals) - np.searchsorted(nvals, nes, side="left"))
            nesPos = len(nnes) - int(np.searchsorted(nnes, 0, side="left"))
            nesHigherAndPos = len(nnes) - int(np.searchsorted(nnes, nes, side="left"))
            # allPos = (nvals >= 0).sum()
            # allHigherAndPos = (nvals >= nes).sum()
            # nesPos = (nnes >=0).sum()
            # nesHigherAndPos = (nnes >= nes).sum()
        else:
            allPos = int(np.searchsorted(nvals, 0, side="left"))
            allHigherAndPos = int(np.searchsorted(nvals, nes, side="right"))
            nesPos = int(np.searchsorted(nnes, 0, side="left"))
            nesHigherAndPos = int(np.searchsorted(nnes, nes, side="right"))
            # allPos = (nvals < 0).sum()
            # allHigherAndPos = (nvals < nes).sum()
            # nesPos = (nnes < 0).sum()
            # nesHigherAndPos = (nnes < nes).sum()

        try:
            pi_norm = allHigherAndPos / float(allPos)
            pi_obs = nesHigherAndPos / float(nesPos)
            fdr = pi_norm / pi_obs
            fdrs.append(fdr if fdr < 1 else 1.0)
        except:
            fdrs.append(1000000000.0)

    logging.debug("Statistical testing finished.............................")

    return fdrs


def gsea_significance(enrichment_scores, enrichment_nulls):
    """Compute nominal pvals, normalized ES, and FDR q value.

    For a given NES(S) = NES* >= 0. The FDR is the ratio of the percentage of all (S,pi) with
    NES(S,pi) >= 0, whose NES(S,pi) >= NES*, divided by the percentage of
    observed S wih NES(S) >= 0, whose NES(S) >= NES*, and similarly if NES(S) = NES* <= 0.
    """
    # For a zero by zero division (undetermined, results in a NaN),
    np.seterr(divide="ignore", invalid="ignore")
    # import warnings
    # warnings.simplefilter("ignore")
    es = np.array(enrichment_scores)
    esnull = np.array(enrichment_nulls)
    logging.debug("Start to compute pvals..................................")
    # P-values.
    pvals = gsea_pval(es, esnull).tolist()

    logging.debug("Start to compute nes and nesnull........................")
    # NES
    nEnrichmentScores, nEnrichmentNulls = normalize(es, esnull)

    logging.debug("Start to compute fdrs..................................")
    # FDR
    fdrs = gsea_fdr(nEnrichmentScores, nEnrichmentNulls)

    # TODO: use multiple testing correction for ssgsea? ssGSEA2.0 use BH correction.
    # https://github.com/broadinstitute/ssGSEA2.0/blob/master/src/ssGSEA2.0.R
    # line 969
    # fdrs, _ = multiple_testing_correction(pvals, alpha=0.05)

    return zip(enrichment_scores, nEnrichmentScores, pvals, fdrs)



from tempfile import TemporaryDirectory, mkdtemp

import numpy as np
import pandas as pd
import pytest

from gseapy.__init__ import enrich, enrichr, gsea, gsva, prerank, replot, ssgsea


@pytest.fixture
def edbDIR():
    return "tests/data"


@pytest.fixture
def genelist():
    return "tests/data/gene_list.txt"


@pytest.fixture
def gseaGCT():
    return "tests/extdata/Leukemia_hgu95av2.trim.txt"


@pytest.fixture
def gseaCLS():
    return "tests/extdata/Leukemia.cls"


@pytest.fixture
def prernk():
    return "tests/data/edb/gsea_data.gsea_data.rnk"


@pytest.fixture
def geneGMT():
    return "tests/data/genes.gmt"


@pytest.fixture
def ssGMT():
    return "tests/data/temp.gmt"


@pytest.fixture
def ssGCT():
    return "tests/extdata/Leukemia_hgu95av2.trim.txt"


@pytest.fixture
def gsvaExpr():
    return "tests/data/expr.gsva.csv"


@pytest.fixture
def gsvaGMT():
    return "tests/data/geneset.gsva.gmt"


def test_gsea(gseaGCT, gseaCLS, geneGMT):
    # Only tests of the command runs successfully,
    # doesnt't check the image
    tmpdir = TemporaryDirectory(dir="tests")
    gs_res = gsea(
        data=gseaGCT,
        gene_sets=geneGMT,
        cls=gseaCLS,
        method="t_test",
        outdir=tmpdir.name,
        permutation_type="phenotype",
        permutation_num=47,
    )
    tmpdir.cleanup()


def test_fdr_gsea(gseaGCT, gseaCLS, geneGMT):
    # Runs and verifies a reasonable result for pval
    # when method="t_test" and permutation_type="gene_set"
    tmpdir = TemporaryDirectory(dir="tests")
    gs_res = gsea(
        data=gseaGCT,
        gene_sets=geneGMT,
        cls=gseaCLS,
        method="t_test",
        outdir=tmpdir.name,
        permutation_type="gene_set",
        permutation_num=47,
        seed=7,
    )
    assert gs_res.res2d["NOM p-val"].max() > 0.0
    tmpdir.cleanup()


def test_prerank(prernk, geneGMT):
    # Only tests of the command runs successfully,
    # doesnt't check the image
    tmpdir = TemporaryDirectory(dir="tests")
    prerank(prernk, geneGMT, tmpdir.name, permutation_num=10)
    tmpdir.cleanup()
    prerank(
        prernk, ["KEGG_2016", "GO_Biological_Process_2021"], None, permutation_num=20
    )


def test_ssgsea1(ssGCT, geneGMT):
    # Only tests of the command runs successfully,
    # doesnt't check the image
    tmpdir = TemporaryDirectory(dir="tests")
    ssgsea(ssGCT, geneGMT, tmpdir.name, permutation_num=100)
    tmpdir.cleanup()


def test_ssgsea2(ssGCT, geneGMT):
    # Only tests of the command runs successfully,
    # doesnt't check the image
    tmpdir = TemporaryDirectory(dir="tests")
    ssgsea(ssGCT, geneGMT, tmpdir.name, permutation_num=0)
    tmpdir.cleanup()
    ssgsea(ssGCT, geneGMT, None, permutation_num=0, correl_norm_type="symrank")
    ssgsea(ssGCT, geneGMT, None, permutation_num=0, correl_norm_type="zscore")


def test_gsva(gsvaExpr, gsvaGMT):
    # Only tests of the command runs successfully,
    # doesnt't check the image
    tmpdir = TemporaryDirectory(dir="tests")
    gsva(gsvaExpr, gsvaGMT, tmpdir.name, kcdf="Gaussian")
    gsva(gsvaExpr, gsvaGMT, tmpdir.name, kcdf="Poisson")
    gsva(gsvaExpr, gsvaGMT, tmpdir.name, kcdf=None)
    tmpdir.cleanup()


def test_gsva_uppercase_conversion():
    """Test GSVA when gene sets are uppercase but expression genes are lowercase.

    Regression test for: 'numpy.float64' object has no attribute 'upper'
    This ensures str() is called before .upper() on gene names.
    """
    np.random.seed(42)
    n_genes = 100
    n_samples = 4
    gene_names = ["gene%d" % i for i in range(n_genes)]
    expr = pd.DataFrame(
        np.random.randn(n_genes, n_samples),
        index=gene_names,
        columns=["s%d" % i for i in range(n_samples)],
    )
    # gene sets with UPPERCASE gene names trigger the _gene_toupper path
    gene_sets = {
        "SET_A": [g.upper() for g in gene_names[:30]],
        "SET_B": [g.upper() for g in gene_names[20:60]],
    }
    tmpdir = TemporaryDirectory(dir="tests")
    result = gsva(data=expr, gene_sets=gene_sets, outdir=tmpdir.name)
    assert result.res2d is not None
    tmpdir.cleanup()


def test_enrichr(genelist, geneGMT):
    # Only tests of the command runs successfully,
    # doesnt't check the image
    tmpdir = TemporaryDirectory(dir="tests")
    enrichr(
        genelist,
        gene_sets=["KEGG_2016", geneGMT],
        background="hsapiens_gene_ensembl",
        outdir=None,
        cutoff=0.5,
    )
    tmpdir.cleanup()
    tmpdir = TemporaryDirectory(dir="tests")
    enrich(genelist, gene_sets=geneGMT, background=None, outdir=tmpdir.name, cutoff=0.5)
    tmpdir.cleanup()


# ---------------------------------------------------------------------------
# Enrichr / enrich — additional test cases
# ---------------------------------------------------------------------------


@pytest.fixture
def small_gene_sets():
    """Small in-memory gene sets dict for local enrichment tests."""
    return {
        "PATHWAY_A": ["IGKC", "CD55", "ABHD4", "PCSK6", "PGD", "ARHGDIB"],
        "PATHWAY_B": ["ITGB2", "CARD6", "MNDA", "ATE1", "LGALS8", "HOMER3"],
        "PATHWAY_C": ["VDAC3", "MAP4K4", "HS2ST1", "FAKE1", "FAKE2", "FAKE3"],
    }


@pytest.fixture
def small_gene_list():
    """A small gene list that partially overlaps with small_gene_sets."""
    return ["IGKC", "CD55", "ABHD4", "ITGB2", "CARD6", "VDAC3", "MAP4K4", "NOTFOUND"]


class TestEnrichLocal:
    """Tests for offline / local enrichment (enrich function with dict gene sets)."""

    def test_enrich_dict_gene_sets(self, small_gene_list, small_gene_sets):
        """enrich() with dict gene sets should return results with expected columns."""
        tmpdir = TemporaryDirectory(dir="tests")
        result = enrich(
            gene_list=small_gene_list,
            gene_sets=small_gene_sets,
            background=None,
            outdir=tmpdir.name,
            cutoff=1.0,
            no_plot=True,
        )
        assert result.res2d is not None
        assert not result.res2d.empty
        expected_cols = {
            "Gene_set",
            "Term",
            "Overlap",
            "P-value",
            "Adjusted P-value",
            "Odds Ratio",
            "Combined Score",
            "Genes",
        }
        assert expected_cols.issubset(set(result.res2d.columns))
        tmpdir.cleanup()

    def test_enrich_returns_all_terms(self, small_gene_list, small_gene_sets):
        """All terms with at least one hit should appear in results."""
        result = enrich(
            gene_list=small_gene_list,
            gene_sets=small_gene_sets,
            background=None,
            outdir=None,
            cutoff=1.0,
            no_plot=True,
        )
        returned_terms = set(result.res2d["Term"])
        # Each pathway has at least some overlap with small_gene_list
        for term in small_gene_sets:
            assert term in returned_terms

    def test_enrich_overlap_format(self, small_gene_list, small_gene_sets):
        """Overlap column should be in 'k/K' format."""
        result = enrich(
            gene_list=small_gene_list,
            gene_sets=small_gene_sets,
            background=None,
            outdir=None,
            cutoff=1.0,
            no_plot=True,
        )
        for overlap in result.res2d["Overlap"]:
            parts = overlap.split("/")
            assert len(parts) == 2
            assert int(parts[0]) > 0
            assert int(parts[1]) > 0
            assert int(parts[0]) <= int(parts[1])

    def test_enrich_pvalue_range(self, small_gene_list, small_gene_sets):
        """P-values and adjusted p-values should be in [0, 1]."""
        result = enrich(
            gene_list=small_gene_list,
            gene_sets=small_gene_sets,
            background=None,
            outdir=None,
            cutoff=1.0,
            no_plot=True,
        )
        assert (result.res2d["P-value"] >= 0).all()
        assert (result.res2d["P-value"] <= 1).all()
        assert (result.res2d["Adjusted P-value"] >= 0).all()
        assert (result.res2d["Adjusted P-value"] <= 1).all()

    def test_enrich_with_integer_background(self, small_gene_list, small_gene_sets):
        """enrich() should work when background is an integer."""
        result = enrich(
            gene_list=small_gene_list,
            gene_sets=small_gene_sets,
            background=20000,
            outdir=None,
            cutoff=1.0,
            no_plot=True,
        )
        assert result.res2d is not None
        assert not result.res2d.empty

    def test_enrich_with_list_background(self, small_gene_list, small_gene_sets):
        """enrich() should work when background is a list of genes."""
        bg = list(
            {g for genes in small_gene_sets.values() for g in genes}
            | set(small_gene_list)
            | {"EXTRA1", "EXTRA2", "EXTRA3", "EXTRA4", "EXTRA5"}
        )
        result = enrich(
            gene_list=small_gene_list,
            gene_sets=small_gene_sets,
            background=bg,
            outdir=None,
            cutoff=1.0,
            no_plot=True,
        )
        assert result.res2d is not None
        assert not result.res2d.empty

    def test_enrich_with_background_file(self, small_gene_list, small_gene_sets):
        """enrich() should work when background is a file path."""
        result = enrich(
            gene_list=small_gene_list,
            gene_sets=small_gene_sets,
            background="tests/data/background.txt",
            outdir=None,
            cutoff=1.0,
            no_plot=True,
        )
        assert result.res2d is not None

    def test_enrich_genes_column_contains_hits(self, small_gene_list, small_gene_sets):
        """Genes column should only contain genes from the input list."""
        result = enrich(
            gene_list=small_gene_list,
            gene_sets=small_gene_sets,
            background=None,
            outdir=None,
            cutoff=1.0,
            no_plot=True,
        )
        input_set = set(small_gene_list)
        for genes_str in result.res2d["Genes"]:
            genes = set(genes_str.split(";"))
            assert genes.issubset(input_set)

    def test_enrich_gmt_file(self, genelist, geneGMT):
        """enrich() should work with a GMT file path."""
        result = enrich(
            gene_list=genelist,
            gene_sets=geneGMT,
            background=None,
            outdir=None,
            cutoff=1.0,
            no_plot=True,
        )
        assert result.res2d is not None

    def test_enrich_no_outdir(self, small_gene_list, small_gene_sets):
        """enrich() should work when outdir=None (no file output)."""
        result = enrich(
            gene_list=small_gene_list,
            gene_sets=small_gene_sets,
            background=None,
            outdir=None,
            cutoff=1.0,
            no_plot=True,
        )
        assert result.res2d is not None

    def test_enrich_cutoff_filters_plot_not_data(
        self, small_gene_list, small_gene_sets
    ):
        """Cutoff should not remove rows from res2d (only affects plotting)."""
        result_strict = enrich(
            gene_list=small_gene_list,
            gene_sets=small_gene_sets,
            background=None,
            outdir=None,
            cutoff=0.0001,
            no_plot=True,
        )
        result_loose = enrich(
            gene_list=small_gene_list,
            gene_sets=small_gene_sets,
            background=None,
            outdir=None,
            cutoff=1.0,
            no_plot=True,
        )
        # Both should contain the same number of terms
        assert len(result_strict.res2d) == len(result_loose.res2d)


class TestEnrichInputFormats:
    """Tests for various input formats accepted by enrich/enrichr."""

    def test_enrich_list_input(self, small_gene_sets):
        """enrich() should accept a plain list of gene names."""
        result = enrich(
            gene_list=["IGKC", "CD55", "ABHD4"],
            gene_sets=small_gene_sets,
            outdir=None,
            cutoff=1.0,
            no_plot=True,
        )
        assert result.res2d is not None

    def test_enrich_series_input(self, small_gene_sets):
        """enrich() should accept a pd.Series of gene names."""
        gene_series = pd.Series(["IGKC", "CD55", "ABHD4", "ITGB2"])
        result = enrich(
            gene_list=gene_series,
            gene_sets=small_gene_sets,
            outdir=None,
            cutoff=1.0,
            no_plot=True,
        )
        assert result.res2d is not None

    def test_enrich_dataframe_input(self, small_gene_sets):
        """enrich() should accept a single-column pd.DataFrame of gene names."""
        gene_df = pd.DataFrame({"genes": ["IGKC", "CD55", "ABHD4", "ITGB2"]})
        result = enrich(
            gene_list=gene_df,
            gene_sets=small_gene_sets,
            outdir=None,
            cutoff=1.0,
            no_plot=True,
        )
        assert result.res2d is not None

    def test_enrich_file_input(self, small_gene_sets):
        """enrich() should accept a file path to a gene list."""
        result = enrich(
            gene_list="tests/data/gene_list.txt",
            gene_sets=small_gene_sets,
            outdir=None,
            cutoff=1.0,
            no_plot=True,
        )
        assert result.res2d is not None

    def test_enrich_comma_separated_gene_sets(self, genelist):
        """enrich() should accept comma-separated GMT file paths."""
        gmt = "tests/data/genes.gmt"
        result = enrich(
            gene_list=genelist,
            gene_sets=gmt,
            outdir=None,
            cutoff=1.0,
            no_plot=True,
        )
        assert result.res2d is not None


class TestEnrichUpperCase:
    """Tests for gene name case handling in local enrichment."""

    def test_enrich_uppercase_gene_sets_lowercase_input(self):
        """When gene sets are UPPERCASE and input is lowercase, should still match."""
        gene_list = ["gene1", "gene2", "gene3", "gene4"]
        gene_sets = {
            "SET_A": ["GENE1", "GENE2", "GENE5", "GENE6"],
            "SET_B": ["GENE3", "GENE4", "GENE7", "GENE8"],
        }
        result = enrich(
            gene_list=gene_list,
            gene_sets=gene_sets,
            outdir=None,
            cutoff=1.0,
            no_plot=True,
        )
        assert result.res2d is not None
        assert not result.res2d.empty

    def test_enrich_mixed_case_query(self):
        """Mixed-case gene lists should still produce results."""
        gene_list = ["Gene1", "GENE2", "gene3"]
        gene_sets = {
            "SET_A": ["GENE1", "GENE2", "GENE3", "GENE4"],
        }
        result = enrich(
            gene_list=gene_list,
            gene_sets=gene_sets,
            outdir=None,
            cutoff=1.0,
            no_plot=True,
        )
        assert result.res2d is not None


class TestEnrichValidation:
    """Tests for input validation and error handling."""

    def test_enrich_empty_gene_list_raises(self, small_gene_sets):
        """enrich() should raise on empty gene list."""
        with pytest.raises(Exception):
            enrich(
                gene_list=[],
                gene_sets=small_gene_sets,
                outdir=None,
                no_plot=True,
            )

    def test_enrich_invalid_gene_sets_type_raises(self):
        """enrich() should raise when gene_sets is an invalid type."""
        with pytest.raises(Exception):
            enrich(
                gene_list=["GENE1", "GENE2"],
                gene_sets=12345,
                outdir=None,
                no_plot=True,
            )

class TestEnrichrAPI:
    """Tests for the EnrichrAPI class directly."""

    def test_enrichr_api_invalid_organism_raises(self):
        from gseapy.enrichr import EnrichrAPI

        with pytest.raises(ValueError, match="Invalid organism"):
            EnrichrAPI(organism="alien")

    def test_enrichr_api_ensure_list_string(self):
        from gseapy.enrichr import EnrichrAPI

        api = EnrichrAPI(organism="human")
        assert api._ensure_list("KEGG_2016") == ["KEGG_2016"]

    def test_enrichr_api_ensure_list_already_list(self):
        from gseapy.enrichr import EnrichrAPI

        api = EnrichrAPI(organism="human")
        assert api._ensure_list(["A", "B"]) == ["A", "B"]

    def test_enrichr_api_organism_aliases(self):
        """Different aliases for the same organism should resolve to the same URL."""
        from gseapy.enrichr import EnrichrAPI

        human_aliases = ["human", "hsapiens", "homo sapiens", "hs"]
        urls = {EnrichrAPI(organism=a).base_url for a in human_aliases}
        assert len(urls) == 1  # all should map to the same URL


class TestEnrichrClass:
    """Tests for the Enrichr class instantiation and helpers."""

    def test_enrichr_class_parse_gmt(self, geneGMT):
        from gseapy.enrichr import Enrichr

        enr = Enrichr(
            gene_list=["A", "B"],
            gene_sets=geneGMT,
            outdir=None,
            no_plot=True,
        )
        gmt = enr._parse_gmt(geneGMT)
        assert isinstance(gmt, dict)
        assert len(gmt) > 0
        for term, genes in gmt.items():
            assert isinstance(term, str)
            assert isinstance(genes, list)
            assert len(genes) > 0
        enr.close()

    def test_enrichr_class_check_uppercase(self):
        from gseapy.enrichr import Enrichr

        enr = Enrichr(
            gene_list=["A", "B"],
            gene_sets={"X": ["A"]},
            outdir=None,
            no_plot=True,
        )
        assert enr.check_uppercase(["BRCA1", "TP53", "EGFR"]) is True
        assert enr.check_uppercase(["brca1", "tp53", "egfr"]) is False
        assert enr.check_uppercase(["123", "456"]) is False  # entrez IDs
        enr.close()

    def test_enrichr_class_is_entrez_id(self):
        from gseapy.enrichr import Enrichr

        enr = Enrichr(
            gene_list=["A"],
            gene_sets={"X": ["A"]},
            outdir=None,
            no_plot=True,
        )
        assert enr._is_entrez_id("12345") is True
        assert enr._is_entrez_id(12345) is True
        assert enr._is_entrez_id("BRCA1") is False
        assert enr._is_entrez_id("") is False
        enr.close()

    def test_enrichr_class_filter_gmt(self):
        from gseapy.enrichr import Enrichr

        enr = Enrichr(
            gene_list=["A"],
            gene_sets={"X": ["A"]},
            outdir=None,
            no_plot=True,
        )
        gmt = {"TERM1": ["A", "B", "C"], "TERM2": ["D", "E"]}
        bg = {"A", "B", "D"}
        filtered = enr.filter_gmt(gmt, bg)
        assert filtered["TERM1"] == ["A", "B"]
        assert filtered["TERM2"] == ["D"]
        enr.close()

    def test_enrichr_class_filter_gmt_removes_empty(self):
        from gseapy.enrichr import Enrichr

        enr = Enrichr(
            gene_list=["A"],
            gene_sets={"X": ["A"]},
            outdir=None,
            no_plot=True,
        )
        gmt = {"TERM1": ["X", "Y"], "TERM2": ["A"]}
        bg = {"A"}
        filtered = enr.filter_gmt(gmt, bg)
        assert "TERM1" not in filtered
        assert filtered["TERM2"] == ["A"]
        enr.close()

    def test_enrichr_class_context_manager(self):
        from gseapy.enrichr import Enrichr

        with Enrichr(
            gene_list=["A", "B"],
            gene_sets={"X": ["A"]},
            outdir=None,
            no_plot=True,
        ) as enr:
            assert enr is not None

    def test_enrichr_class_parse_genelists_list(self):
        from gseapy.enrichr import Enrichr

        enr = Enrichr(
            gene_list=["BRCA1", "TP53", "EGFR"],
            gene_sets={"X": ["BRCA1"]},
            outdir=None,
            no_plot=True,
        )
        result = enr.parse_genelists()
        assert "BRCA1" in result
        assert "TP53" in result
        assert enr._isezid is False
        enr.close()

    def test_enrichr_class_parse_genelists_entrez(self):
        from gseapy.enrichr import Enrichr

        enr = Enrichr(
            gene_list=["672", "7157", "1956"],
            gene_sets={"X": ["672"]},
            outdir=None,
            no_plot=True,
        )
        result = enr.parse_genelists()
        assert enr._isezid is True
        assert isinstance(enr._gls, set)
        enr.close()

    def test_enrichr_class_parse_background_none(self):
        from gseapy.enrichr import Enrichr

        enr = Enrichr(
            gene_list=["A"],
            gene_sets={"X": ["A"]},
            outdir=None,
            background=None,
            no_plot=True,
        )
        gmt = {"TERM1": ["A", "B", "C"]}
        bg = enr.parse_background(gmt)
        assert isinstance(bg, set)
        assert bg == {"A", "B", "C"}
        enr.close()

    def test_enrichr_class_parse_background_int(self):
        from gseapy.enrichr import Enrichr

        enr = Enrichr(
            gene_list=["A"],
            gene_sets={"X": ["A"]},
            outdir=None,
            background=20000,
            no_plot=True,
        )
        bg = enr.parse_background()
        assert bg == 20000
        enr.close()

    def test_enrichr_class_parse_background_set(self):
        from gseapy.enrichr import Enrichr

        bg_genes = ["A", "B", "C", "D"]
        enr = Enrichr(
            gene_list=["A"],
            gene_sets={"X": ["A"]},
            outdir=None,
            background=bg_genes,
            no_plot=True,
        )
        bg = enr.parse_background()
        assert isinstance(bg, set)
        assert bg == {"A", "B", "C", "D"}
        enr.close()


class TestEnrichOnline:
    """Tests for Enrichr online API calls (these require network access)."""

    @pytest.mark.network
    def test_enrichr_kegg(self, genelist):
        """enrichr() with KEGG_2016 library should return results."""
        result = enrichr(
            gene_list=genelist,
            gene_sets="KEGG_2016",
            organism="human",
            outdir=None,
            cutoff=0.5,
            no_plot=True,
        )
        assert result.res2d is not None
        assert not result.res2d.empty

    @pytest.mark.network
    def test_enrichr_multiple_libraries(self, genelist):
        """enrichr() should handle multiple Enrichr library names."""
        result = enrichr(
            gene_list=genelist,
            gene_sets=["KEGG_2016", "GO_Biological_Process_2021"],
            organism="human",
            outdir=None,
            cutoff=0.5,
            no_plot=True,
        )
        assert result.res2d is not None
        assert "Gene_set" in result.res2d.columns


def test_replot(edbDIR):
    # Only tests of the command runs successfully,
    # doesnt't check the image
    tmpdir = TemporaryDirectory(dir="tests")
    replot(edbDIR, tmpdir.name)
    tmpdir.cleanup()



#! python
# -*- coding: utf-8 -*-

import json
import logging
import os
from typing import Any, Dict, Iterable, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
from pandas.api.types import is_object_dtype, is_string_dtype

from gseapy.plot import GSEAPlot, TracePlot, gseaplot, heatmap
from gseapy.utils import DEFAULT_CACHE_PATH, log_init, mkdirs, retry


class GMT:
    """A collection of gene set dictionaries with metadata.

    Attributes:
        _collections: Dict[str, Dict[str, Any]] - Stores gene set collections
            key: collection name
            value: {
                'genes': Dict[str, List[str]] - Gene set mappings
                'description': str - Collection description
                'source': str - Source of the gene sets
            }
    """

    def __init__(
        self,
        mapping: Optional[Dict[str, List[str]]] = None,
        description: Optional[str] = None,
        source: Optional[str] = None,
        name: Optional[str] = "default",
    ):
        """Initialize a GMT collection.

        Args:
            mapping: Initial gene set dictionary
            description: Description of the gene sets
            source: Source of the gene sets
            name: Name of this collection
        """
        self._collections = {}
        if mapping is not None:
            self.add(mapping, description, source, name)

    def add(
        self,
        mapping: Dict[str, List[str]],
        description: Optional[str] = None,
        source: Optional[str] = None,
        name: Optional[str] = "default",
    ):
        """Add a gene set collection with metadata.

        Args:
            mapping: Gene set dictionary to add
            description: Description of the gene sets
            source: Source of the gene sets
            name: Name for this collection
        """
        self._collections[name] = {
            "genes": mapping,
            "description": description,
            "source": source,
        }

    def get(self, name: str = "default") -> Dict[str, List[str]]:
        """Get gene sets by collection name."""
        return self._collections[name]["genes"]

    def get_metadata(self, name: str = "default") -> Dict[str, Any]:
        """Get metadata for a collection."""
        collection = self._collections[name]
        return {
            "description": collection["description"],
            "source": collection["source"],
        }

    def write(self, ofname: str):
        """Write GMT file to disk."""
        with open(ofname, "w") as out:
            for name, collection in self._collections.items():
                for key, genes in collection["genes"].items():
                    desc = collection["description"] or ""
                    line = [key, desc] + genes
                    out.write("\t".join(line) + "\n")

    def filter(
        self,
        min_size: Optional[int] = None,
        max_size: Optional[int] = None,
        gene_list: Optional[List[str]] = None,
        collections: Optional[List[str]] = None,
    ) -> "GMT":
        """Filter gene sets based on size and gene membership.

        Args:
            min_size: Minimum number of genes in a set
            max_size: Maximum number of genes in a set
            gene_list: Only keep genes present in this list
            collections: Only keep these named collections

        Returns:
            A new filtered GMT object
        """
        filtered = GMT()

        # Filter collections if specified
        colls = collections or self._collections.keys()

        for name in colls:
            if name not in self._collections:
                continue

            collection = self._collections[name]
            filtered_mapping = {}

            for term, genes in collection["genes"].items():
                # Filter genes if gene_list provided
                if gene_list is not None:
                    genes = [g for g in genes if g in gene_list]

                # Apply size filters
                if min_size is not None and len(genes) < min_size:
                    continue
                if max_size is not None and len(genes) > max_size:
                    continue

                # Only add if genes remain after filtering
                if genes:
                    filtered_mapping[term] = genes

            # Only add collection if it has gene sets after filtering
            if filtered_mapping:
                filtered.add(
                    filtered_mapping,
                    description=collection["description"],
                    source=collection["source"],
                    name=name,
                )

        return filtered

    @classmethod
    def read(cls, paths: str, source: Optional[str] = None) -> "GMT":
        """Read GMT files into a collection.

        Args:
            paths: Comma-separated list of GMT file paths
            source: Source annotation for the files
        """
        gmt = cls()
        for path in paths.strip().split(","):
            name = os.path.basename(path)
            mapping = {}
            with open(path) as f:
                for line in f:
                    items = line.strip().split("\t")
                    key = items[0]
                    desc = items[1]
                    genes = items[2:]
                    mapping[key] = genes
            gmt.add(mapping, desc, source, name)
        return gmt

    def __getitem__(self, name: str) -> Dict[str, List[str]]:
        """Dictionary-like access to gene sets."""
        return self.get(name)

    def __iter__(self):
        """Iterate over collection names."""
        return iter(self._collections)

    def items(self):
        """Iterate over (name, gene_sets) pairs."""
        return ((name, coll["genes"]) for name, coll in self._collections.items())


class GSEAbase(object):
    """base class of GSEA."""

    def __init__(
        self,
        outdir: Optional[str] = None,
        gene_sets: Union[List[str], str, Dict[str, str]] = "KEGG_2016",
        module: str = "base",
        threads: int = 1,
        enrichr_url: str = "http://maayanlab.cloud",
        verbose: bool = False,
    ):
        self.outdir = outdir
        self.gene_sets = gene_sets
        self.fdr = 0.05
        self.module = module
        self.res2d = None
        self.ranking = None
        self.ascending = False
        self.verbose = verbose
        self._threads = threads
        self.ENRICHR_URL = enrichr_url
        self.pheno_pos = ""
        self.pheno_neg = ""
        self.permutation_num = 0
        self._LIBRARY_LIST_URL = "https://maayanlab.cloud/speedrichr/api/listlibs"
        self._gene_isupper = True
        self._gene_toupper = False
        self._set_cores()
        # init logger
        self.prepare_outdir()

    def __del__(self):
        if hasattr(self, "_logger"):
            handlers = self._logger.handlers[:]
            for handler in handlers:
                handler.close()  # close file
                self._logger.removeHandler(handler)

    def prepare_outdir(self):
        """create temp directory."""
        self._outdir = self.outdir
        logfile = None
        if isinstance(self.outdir, str):
            mkdirs(self.outdir)
            logfile = os.path.join(
                self.outdir, "gseapy.%s.%s.log" % (self.module, id(self))
            )
        self._logfile = logfile
        self._logger = log_init(
            name=str(self.module) + str(id(self)),
            log_level=logging.INFO if self.verbose else logging.WARNING,
            filename=logfile,
        )

    def _set_cores(self):
        """set cpu numbers to be used"""

        cpu_num = os.cpu_count() - 1
        if self._threads > cpu_num:
            cores = cpu_num
        elif self._threads < 1:
            cores = 1
        else:
            cores = self._threads
        # have to be int if user input is float
        self._threads = int(cores)

    def _read_file(self, path: str) -> pd.DataFrame:
        """
        read file, and return dataframe (first column are gene IDs)
        """
        # just txt file like input
        header, sep = "infer", "\t"
        # GCT input format?
        if path.endswith(".gct"):
            rank_metric = pd.read_csv(
                path, skiprows=1, comment="#", index_col=0, sep=sep
            )
        else:
            if path.endswith(".csv"):
                sep = ","
            if path.endswith(".rnk"):
                header = None
            rank_metric = pd.read_csv(
                path, comment="#", index_col=0, sep=sep, header=header
            )
            if rank_metric.shape[1] == 1:
                # rnk file like input
                rank_metric.columns = rank_metric.columns.astype(str)

        return rank_metric.select_dtypes(include=[np.number]).reset_index()

    def _reset_index(self, rank_metric: pd.DataFrame):
        """
        helper function to reset index if index is already gene_names
        """
        # handle index is already gene_names
        # pandas 3.0 compatibility: check for both object and string dtypes
        if is_object_dtype(rank_metric.index.dtype) or is_string_dtype(
            rank_metric.index.dtype
        ):
            # Try to check if all elements can be converted to numbers
            try:
                # is_string_numbers = True, don't reset index
                pd.to_numeric(rank_metric.index)
            except (ValueError, TypeError):
                # Contains non-numeric strings, likely gene names
                # is_string_numbers = False
                rank_metric = rank_metric.reset_index()
        return rank_metric

    def _load_data(self, exprs: Union[str, pd.Series, pd.DataFrame]) -> pd.DataFrame:
        """
        helper function to read data
        """
        # load data
        if isinstance(exprs, pd.DataFrame):
            rank_metric = exprs.copy()
            # handle dataframe with gene_name as index.
            self._logger.debug("Input data is a DataFrame with gene names")
            # handle index is already gene_names
            rank_metric = self._reset_index(rank_metric)
            # pandas 3.0 compatibility: check for both object and string dtypes
            if not (
                is_object_dtype(rank_metric.columns.dtype)
                or is_string_dtype(rank_metric.columns.dtype)
            ):
                rank_metric.columns = rank_metric.columns.astype(str)

        elif isinstance(exprs, pd.Series):
            # change to DataFrame
            self._logger.debug("Input data is a Series with gene names")
            if not isinstance(exprs.name, str):
                if exprs.name is None:
                    # rename col if name attr is none
                    exprs.name = "sample1"
                elif hasattr(exprs.name, "dtype"):
                    # pandas 3.0 compatibility: check for both object and string dtypes
                    if not (
                        is_object_dtype(exprs.name.dtype)
                        or is_string_dtype(exprs.name.dtype)
                    ):
                        exprs.name = exprs.name.astype(str)
                else:
                    exprs.name = str(exprs.name)
            rank_metric = exprs.reset_index()
        elif os.path.isfile(exprs):
            rank_metric = self._read_file(exprs)

        else:
            raise Exception("Error parsing expression values!")
        # select numbers
        # rank_metric = rank_metric.select_dtypes(include=[np.number])
        return rank_metric

    def _check_data(self, exprs: pd.DataFrame) -> pd.DataFrame:
        """
        check NAs, duplicates, and select number columns
        exprs: dataframe, the frist column must be gene identifiers

        return: dataframe, index is gene ids
        """
        ## if gene names contain NA, drop them
        if exprs.iloc[:, 0].isnull().any():
            exprs.dropna(subset=[exprs.columns[0]])
        ## then fill na for numeric columns
        if exprs.isnull().any().sum() > 0:
            self._logger.warning("Input data contains NA, filled NA with 0")
            exprs.dropna(how="all", inplace=True)  # drop rows with all NAs
            exprs = exprs.fillna(0)
        ## check duplicated IDs
        # set gene name as index
        exprs.set_index(keys=exprs.columns[0], inplace=True)
        # select numberic columns
        df = exprs.select_dtypes(include=[np.number])
        # microarray data may contained multiple probs of same gene, average them
        if df.index.duplicated().sum() > 0:
            self._logger.warning(
                "Found duplicated gene names, values averaged by gene names!"
            )
            df = df.groupby(level=0).mean()
        # check whether contains infinity values
        if np.isinf(df).values.sum() > 0:
            self._logger.warning("Input gene rankings contains inf values!")
            col_min_max = {
                np.inf: df[np.isfinite(df)].max(),  # column-wise max
                -np.inf: df[np.isfinite(df)].min(),  # column-wise min
            }
            df = df.replace({col: col_min_max for col in df.columns})
        return df

    def _is_entrez_id(self, idx: Union[int, str]) -> bool:
        """
        Check if an index is Entrez ID.

        Parameters
        ----------
        idx : str or int
            An index to be checked.

        Returns
        -------
        bool
            Whether the index is Entrez ID.
        """
        try:
            int(idx)
            return True
        except:
            return False

    def check_uppercase(self, gene_list: List[Union[str, int]]) -> bool:
        """
        Check whether a list of gene names are mostly in uppercase.

        Parameters
        ----------
        gene_list : list, int
            A list of gene names or Entrez IDs

        Returns
        -------
        bool
            Whether the list of gene names are mostly in uppercase
        """
        # if all gene names are Entrez IDs, don't check uppercase
        if all([self._is_entrez_id(g) for g in gene_list]):
            return False
        is_upper = [str(s).isupper() for s in gene_list]
        if sum(is_upper) / len(is_upper) >= 0.9:
            return True
        return False

    def make_unique(self, rank_metric: pd.DataFrame, col_idx: int) -> pd.DataFrame:
        """
        make gene id column unique by adding a digit, similar to R's make.unique
        """
        id_col = rank_metric.columns[col_idx]
        if rank_metric.duplicated(subset=id_col).sum() > 0:
            self._logger.info("Input gene rankings contains duplicated IDs")
            mask = rank_metric.duplicated(subset=id_col, keep=False)
            dups = (
                rank_metric.loc[mask, id_col]
                .to_frame()
                .groupby(id_col)
                .cumcount()
                .map(lambda c: "_" + str(c) if c else "")
            )
            rank_metric.loc[mask, id_col] = rank_metric.loc[mask, id_col] + dups
        return rank_metric

    def load_gmt_only(
        self, gmt: Union[List[str], str, Dict[str, str]]
    ) -> Dict[str, List[str]]:
        """parse gene_sets.
        gmt: List, Dict, Strings

        However,this function will merge different gene sets into one big dict to
        save computation time for later.
        """
        genesets_dict = dict()

        if isinstance(gmt, dict):
            genesets_dict = gmt.copy()
        elif isinstance(gmt, str):
            gmts = gmt.split(",")
            if len(gmts) > 1:
                for gm in gmts:
                    tdt = self.parse_gmt(gm)
                    for k, v in tdt.items():
                        new_k = os.path.split(gm)[-1] + "__" + k
                        genesets_dict[new_k] = v
            else:
                genesets_dict = self.parse_gmt(gmt)
        elif isinstance(gmt, list):
            for i, gm in enumerate(gmt):
                prefix = str(i)
                if isinstance(gm, dict):
                    tdt = gm.copy()
                elif isinstance(gm, str):
                    tdt = self.parse_gmt(gm)
                    prefix = os.path.split(gm)[-1]
                else:
                    continue
                for k, v in tdt.items():
                    new_k = prefix + "__" + k
                    genesets_dict[new_k] = v
        else:
            raise Exception("Error parsing gmt parameter for gene sets")

        if len(genesets_dict) == 0:
            raise Exception("Error parsing gmt parameter for gene sets")
        return genesets_dict

    def load_gmt(
        self, gene_list: Iterable[str], gmt: Union[List[str], str, Dict[str, str]]
    ) -> Dict[str, List[str]]:
        """load gene set dict"""

        genesets_dict = self.load_gmt_only(gmt)
        subsets = list(genesets_dict.keys())
        if not subsets:  # Check if empty
            raise ValueError("Empty gene sets dictionary")
        entry1st = genesets_dict[subsets[0]]
        # gmt_is_entrez = all([self._is_entrez_id(x) for x in entry1st])
        # gene_is_entrez = all([self._is_entrez_id(x) for x in gene_list])
        gene_dict = {g: i for i, g in enumerate(gene_list)}
        # Check uppercase for up to 20 sets
        sample_size = min(20, len(subsets))
        ups = []
        for s in subsets[:sample_size]:
            ups.append(self.check_uppercase(genesets_dict[s]))

        if (not self._gene_isupper) and all(ups):
            # set flag to True, means use uppercase version of gene symbols
            self._gene_toupper = True
            gene_dict_upper = {str(g).upper(): i for i, g in enumerate(gene_list)}

        # filter gene sets
        for subset in subsets:
            subset_list = set(genesets_dict.get(subset))  # remove duplicates
            # drop genes not found in the gene list
            if (not self._gene_isupper) and all(ups):
                gene_overlap = [g for g in subset_list if g in gene_dict_upper]
            else:
                gene_overlap = [g for g in subset_list if g in gene_dict]

            tag_len = len(gene_overlap)
            if (self.min_size <= tag_len <= self.max_size) and tag_len < len(gene_list):
                # tag_len should < gene_list
                genesets_dict[subset] = gene_overlap
                continue
            del genesets_dict[subset]

        filsets_num = len(subsets) - len(genesets_dict)
        self._logger.info(
            "%04d gene_sets have been filtered out when max_size=%s and min_size=%s"
            % (filsets_num, self.max_size, self.min_size)
        )

        if filsets_num == len(subsets):
            msg = (
                "No gene sets passed through filtering condition !!! \n"
                + "Hint 1: Try to lower min_size or increase max_size !\n"
                + "Hint 2: Check gene symbols are identifiable to your gmt input.\n"
                + "Hint 3: Gene symbols curated in Enrichr web services are all upcases.\n"
            )
            self._logger.error(msg)
            dict_head = "{ %s: [%s]}" % (subsets[0], ", ".join(entry1st))
            self._logger.error(
                "The first entry of your gene_sets (gmt) look like this : %s"
                % dict_head
            )
            self._logger.error(
                "The first 5 genes look like this : [ %s ]"
                % (", ".join(list(gene_list)[:5]))
            )
            raise LookupError(msg)

        # self._gmtdct = genesets_dict
        return genesets_dict

    def parse_gmt(self, gmt: str) -> Dict[str, List[str]]:
        """gmt parser when input is a string"""

        if gmt.lower().endswith(".gmt"):
            genesets_dict = {}
            with open(gmt) as genesets:
                for line in genesets:
                    entries = line.strip().split("\t")
                    key = entries[0]
                    genes = [g.split(",")[0] for g in entries[2:]]
                    genesets_dict[key] = genes
            return genesets_dict

        tmpname = "Enrichr." + gmt + ".gmt"
        tempath = os.path.join(DEFAULT_CACHE_PATH, tmpname)
        # if file already download
        if os.path.isfile(tempath):
            self._logger.info(
                "Enrichr library gene sets already downloaded in: %s, use local file"
                % DEFAULT_CACHE_PATH
            )
            return self.parse_gmt(tempath)

        elif gmt in self.get_libraries():
            return self._download_libraries(gmt)
        else:
            self._logger.error("No supported gene_sets: %s" % gmt)
        return dict()

    def get_libraries(self) -> List[str]:
        """return active enrichr library name.Offical API"""

        lib_url = self.ENRICHR_URL + "/Enrichr/datasetStatistics"
        # "LIBRARY_LIST_URL": "https://maayanlab.cloud/speedrichr/api/listlibs",
        s = retry(num=5)
        response = s.get(lib_url, verify=True)
        if not response.ok:
            raise Exception("Error getting the Enrichr libraries")
        libs_json = json.loads(response.text)
        libs = [lib["libraryName"] for lib in libs_json["statistics"]]

        return sorted(libs)

    def _download_libraries(self, libname: str) -> Dict[str, List[str]]:
        """Download enrichr libraries. Only Support Enrichr libraries now"""
        self._logger.info("Downloading and generating Enrichr library gene sets......")
        s = retry(5)
        # queery string
        ENRICHR_URL = self.ENRICHR_URL + "/Enrichr/geneSetLibrary"
        query_string = "?mode=text&libraryName=%s"
        # get
        response = s.get(
            ENRICHR_URL + query_string % libname, timeout=None, stream=True
        )
        if not response.ok:
            raise Exception(
                "Error fetching gene set library, input name is correct for the organism you've set?."
            )
        # reformat to dict and save to disk
        mkdirs(DEFAULT_CACHE_PATH)
        genesets_dict = {}
        outname = "Enrichr.%s.gmt" % libname  # pattern: database.library.gmt
        gmtout = open(os.path.join(DEFAULT_CACHE_PATH, outname), "w")
        for line in response.iter_lines(chunk_size=1024, decode_unicode="utf-8"):
            line = line.strip().split("\t")
            k = line[0]
            v = map(lambda x: x.split(",")[0], line[2:])
            v = list(filter(lambda x: True if len(x) else False, v))
            genesets_dict[k] = v
            outline = "%s\t%s\t%s\n" % (k, line[1], "\t".join(v))
            gmtout.write(outline)
        gmtout.close()

        return genesets_dict

    def _heatmat(self, df: pd.DataFrame, classes: List[str]):
        """only use for gsea heatmap"""

        cls_booA = list(map(lambda x: True if x == self.pheno_pos else False, classes))
        cls_booB = list(map(lambda x: True if x == self.pheno_neg else False, classes))
        datA = df.loc[:, cls_booA]
        datB = df.loc[:, cls_booB]
        datAB = pd.concat([datA, datB], axis=1)
        self.heatmat = datAB
        return

    def _plotting(self, metric: Dict[str, Union[pd.Series, pd.DataFrame]]):
        """Plotting API.
        :param metric: sorted pd.Series with rankings values.
        """

        # no values need to be returned
        if self._outdir is None:
            return
        # indices = self.res2d["NES"].abs().sort_values(ascending=False).index
        indices = self.res2d.index
        # Plotting
        for i, idx in enumerate(indices):
            record = self.res2d.iloc[idx]
            if self.module != "ssgsea" and record["FDR q-val"] > 0.25:
                continue
            if i >= self.graph_num:  # already sorted by abs(NES) in descending order
                break
            # if self.res2d["Name"].nunique() > 1 and hasattr(
            #     self, "_metric_dict"
            # ):  # self.module != "ssgsea":
            #     key = record["Name"]
            #     rank_metric = metric[key]
            #     hit = self._results[key][record["Term"]]["hits"]
            #     RES = self._results[key][record["Term"]]["RES"]
            # else:
            #     rank_metric = metric[self.module]
            #     hit = self._results[record["Term"]]["hits"]
            #     RES = self._results[record["Term"]]["RES"]
            key = record["Name"]
            rank_metric = metric[key]
            hit = self._results[key][record["Term"]]["hits"]
            RES = self._results[key][record["Term"]]["RES"]

            outdir = os.path.join(self.outdir, record["Name"])
            mkdirs(outdir)
            term = record["Term"].replace("/", "-").replace(":", "_")
            outfile = os.path.join(outdir, "{0}.{1}".format(term, self.format))
            if self.module == "gsea":
                outfile2 = "{0}/{1}.heatmap.{2}".format(outdir, term, self.format)
                heatmat = self.heatmat.iloc[hit, :]
                width = np.clip(heatmat.shape[1], 4, 20)
                height = np.clip(heatmat.shape[0], 4, 20)
                heatmap(
                    df=heatmat,
                    title=record["Term"].split("__")[-1],
                    ofname=outfile2,
                    z_score=0,
                    figsize=(width, height),
                    xticklabels=True,
                    yticklabels=True,
                )

            if self.permutation_num > 0:
                # skip plotting when nperm=0
                gseaplot(
                    term=record["Term"].split("__")[-1],
                    hits=hit,
                    nes=record["NES"],
                    pval=record["NOM p-val"],
                    fdr=record["FDR q-val"],
                    RES=RES,
                    rank_metric=rank_metric,
                    pheno_pos=self.pheno_pos,
                    pheno_neg=self.pheno_neg,
                    figsize=self.figsize,
                    ofname=outfile,
                )

    def _to_df(
        self,
        gsea_summary: List[Dict],
        gmt: Dict[str, List[str]],
        metric: Dict[str, pd.Series],
    ) -> pd.DataFrame:
        """Convernt GSEASummary to DataFrame

        rank_metric: Must be sorted in descending order already
        """

        res_df = pd.DataFrame(
            index=range(len(gsea_summary)),
            columns=[
                "name",
                "term",
                "es",
                "nes",
                "pval",
                "fdr",
                "fwerp",
                "tag %",
                "gene %",
                "lead_genes",
                "matched_genes",
                "hits",
                "RES",
            ],
        )
        # res = OrderedDict()

        for i, gs in enumerate(gsea_summary):
            # reformat gene list.
            name = (
                self._metric_dict[str(gs.index)]
                if (gs.index is not None)
                else self.module
            )
            _genes = metric[name].index.values[gs.hits]
            genes = ";".join([str(g).strip() for g in _genes])
            RES = np.array(gs.run_es)
            lead_genes = ""
            tag_frac = ""
            gene_frac = ""
            if len(RES) > 1:
                # extract leading edge genes
                if float(gs.es) >= 0:
                    # RES -> ndarray, ind -> list
                    es_i = RES.argmax()
                    ldg_pos = list(filter(lambda x: x <= es_i, gs.hits))
                    gene_frac = (es_i + 1) / len(metric[name])
                else:
                    es_i = RES.argmin()
                    ldg_pos = list(filter(lambda x: x >= es_i, gs.hits))
                    ldg_pos.reverse()
                    gene_frac = (len(metric[name]) - es_i) / len(metric[name])

                # tag_frac = len(ldg_pos) / len(gmt[gs.term])
                gene_frac = "{0:.2%}".format(gene_frac)
                lead_genes = ";".join(list(map(str, metric[name].iloc[ldg_pos].index)))
                tag_frac = "%s/%s" % (len(ldg_pos), len(gmt[gs.term]))

            e = pd.Series(
                [
                    name,
                    gs.term,
                    gs.es,
                    gs.nes,
                    gs.pval,
                    gs.fdr,
                    gs.fwerp,
                    tag_frac,
                    gene_frac,
                    lead_genes,
                    genes,
                    gs.hits,
                    gs.run_es,
                ],
                index=res_df.columns,
            )
            res_df.iloc[i, :] = e
        return res_df

    def to_df(
        self,
        gsea_summary: List[Dict],
        gmt: Dict[str, List[str]],
        rank_metric: Union[pd.Series, pd.DataFrame],
        indices: Optional[List] = None,
    ):
        """Convernt GSEASummary to DataFrame

        rank_metric: if a Series, then it must be sorted in descending order already
                     if a DataFrame, indices must not None.
        indices: Only works for DataFrame input. Stores the indices of sorted array
        """
        if isinstance(rank_metric, pd.DataFrame) and (indices is not None):
            self._metric_dict = {str(c): n for c, n in enumerate(rank_metric.columns)}
            metric = {
                n: rank_metric.iloc[indices[i], i]  # series with gene_name indexed
                for i, n in enumerate(rank_metric.columns)  # indices is a 2d list
            }
        else:
            metric = {self.module: rank_metric}
            self._metric_dict = {self.module: self.module}

        res_df = self._to_df(gsea_summary, gmt, metric)
        self._results = {}
        # save dict
        # if res_df["name"].nunique() >= 2:
        #     for name, dd in res_df.groupby(["name"]):
        #         self._results[name] = dd.set_index("term").to_dict(orient="index")
        # else:
        #     self._results = res_df.set_index("term").to_dict(orient="index")
        for name, dd in res_df.groupby("name"):
            self._results[name] = dd.set_index("term").to_dict(orient="index")
        # trim
        res_df.rename(
            columns={
                "name": "Name",
                "term": "Term",
                "es": "ES",
                "nes": "NES",
                "pval": "NOM p-val",
                "fdr": "FDR q-val",
                "fwerp": "FWER p-val",
                "tag %": "Tag %",
                "gene %": "Gene %",
                "lead_genes": "Lead_genes",
            },
            inplace=True,
        )
        # res_df["Gene %"] = res_df["Gene %"].map(lambda x: "{0:.2%}".format(x) if x !="" else "")
        # trim
        dc = ["RES", "hits", "matched_genes"]
        if self.permutation_num == 0:
            dc += [
                "NOM p-val",
                "FWER p-val",
                "FDR q-val",
                "Tag %",
                "Gene %",
                "Lead_genes",
            ]
        if self.module == "gsva":
            dc += ["NES"]
        # re-order by NES
        # for pandas > 1.1, use df.sort_values(by='B', key=abs) will sort by abs value
        self.res2d = res_df.reindex(
            res_df["NES"].abs().sort_values(ascending=False).index
        ).reset_index(drop=True)
        self.res2d.drop(dc, axis=1, inplace=True)

        if self._outdir is not None:
            out = os.path.join(
                self.outdir,
                "gseapy.{b}.{c}.report.csv".format(
                    b=self.permutation_type, c=self.module
                ),
            )
            self.res2d.to_csv(out, index=False, float_format="%.6e")
            with open(os.path.join(self.outdir, "gene_sets.gmt"), "w") as gout:
                for term, genes in gmt.items():
                    collection = ""
                    if term.find("__") > -1:
                        collections = term.split("__")
                        collection = collections[0]
                        term = collections[1]
                    gg = "\t".join(genes)
                    gout.write(f"{term}\t{collection}\t{gg}\n")
        # generate gseaplots
        if not self._noplot:
            self._plotting(metric)
        return

    @property
    def results(self):
        """
        compatible to old style
        """
        keys = list(self._results.keys())
        if len(keys) == 1:
            return self._results[keys[0]]
        return self._results

    def enrichment_score(
        self,
        gene_list: Iterable[str],
        correl_vector: Iterable[float],
        gene_set: Dict[str, List[str]],
        weight: float = 1.0,
        nperm: int = 1000,
        seed: int = 123,
        single: bool = False,
        scale: bool = False,
    ):
        """This is the most important function of GSEApy. It has the same algorithm with GSEA and ssGSEA.

        :param gene_list:       The ordered gene list gene_name_list, rank_metric.index.values
        :param gene_set:        gene_sets in gmt file, please use gmt_parser to get gene_set.
        :param weight:  It's the same with gsea's weighted_score method. Weighting by the correlation
                                is a very reasonable choice that allows significant gene sets with less than perfect coherence.
                                options: 0(classic),1,1.5,2. default:1. if one is interested in penalizing sets for lack of
                                coherence or to discover sets with any type of nonrandom distribution of tags, a value p < 1
                                might be appropriate. On the other hand, if one uses sets with large number of genes and only
                                a small subset of those is expected to be coherent, then one could consider using p > 1.
                                Our recommendation is to use p = 1 and use other settings only if you are very experienced
                                with the method and its behavior.

        :param correl_vector:   A vector with the correlations (e.g. signal to noise scores) corresponding to the genes in
                                the gene list. Or rankings, rank_metric.values
        :param nperm:           Only use this parameter when computing esnull for statistical testing. Set the esnull value
                                equal to the permutation number.
        :param seed:            Random state for initializing gene list shuffling. Default: seed=None

        :return:

        ES: Enrichment score (real number between -1 and +1)

        ESNULL: Enrichment score calculated from random permutations.

        Hits_Indices: Index of a gene in gene_list, if gene is included in gene_set.

        RES: Numerical vector containing the running enrichment score for all locations in the gene list .

        """
        N = len(gene_list)
        # Test whether each element of a 1-D array is also present in a second array
        # It's more intuitive here than original enrichment_score source code.
        # use .astype to covert bool to integer
        tag_indicator = np.isin(gene_list, gene_set, assume_unique=True).astype(
            int
        )  # notice that the sign is 0 (no tag) or 1 (tag)

        if weight == 0:
            correl_vector = np.repeat(1, N)
        else:
            correl_vector = np.abs(correl_vector) ** weight

        # get indices of tag_indicator
        hit_ind = np.flatnonzero(tag_indicator).tolist()
        # if used for compute esnull, set esnull equal to permutation number, e.g. 1000
        # else just compute enrichment scores
        # set axis to 1, because we have 2D array
        axis = 1
        tag_indicator = np.tile(tag_indicator, (nperm + 1, 1))
        correl_vector = np.tile(correl_vector, (nperm + 1, 1))
        # gene list permutation
        rs = np.random.RandomState(seed)
        for i in range(nperm):
            rs.shuffle(tag_indicator[i])
        # np.apply_along_axis(rs.shuffle, 1, tag_indicator)

        Nhint = tag_indicator.sum(axis=axis, keepdims=True)
        sum_correl_tag = np.sum(correl_vector * tag_indicator, axis=axis, keepdims=True)
        # compute ES score, the code below is identical to gsea enrichment_score method.
        no_tag_indicator = 1 - tag_indicator
        Nmiss = N - Nhint
        norm_tag = 1.0 / sum_correl_tag
        norm_no_tag = 1.0 / Nmiss

        RES = np.cumsum(
            tag_indicator * correl_vector * norm_tag - no_tag_indicator * norm_no_tag,
            axis=axis,
        )

        if scale:
            RES = RES / N
        if single:
            es_vec = RES.sum(axis=axis)
        else:
            max_ES, min_ES = RES.max(axis=axis), RES.min(axis=axis)
            es_vec = np.where(np.abs(max_ES) > np.abs(min_ES), max_ES, min_ES)
        # extract values
        es, esnull, RES = es_vec[-1], es_vec[:-1], RES[-1, :]

        return es, esnull, hit_ind, RES

    def plot(
        self,
        terms: Union[str, List[str]],
        colors: Optional[Union[str, List[str]]] = None,
        legend_kws: Optional[Dict[str, Any]] = None,
        figsize: Tuple[float, float] = (4, 5),
        show_ranking: bool = True,
        ofname: Optional[str] = None,
    ):
        """
        terms: str, list.  terms/pathways to show
        colors: str, list. list of colors for each term/pathway
        legend_kws: kwargs to pass to ax.legend. e.g. `loc`, `bbox_to_achor`.
        ofname: savefig
        """
        # if hasattr(self, "results"):
        if self.module in ["ssgsea", "gsva"]:
            raise NotImplementedError("not for ssgsea")
        keys = list(self._results.keys())
        if len(keys) > 1:
            raise NotImplementedError("Multiple Dataset input No supported yet!")

        ranking = self.ranking if show_ranking else None
        if isinstance(terms, str):
            gsdict = self.results[terms]
            g = GSEAPlot(
                term=terms,
                tag=gsdict["hits"],
                rank_metric=ranking,
                runes=gsdict["RES"],
                nes=gsdict["nes"],
                pval=gsdict["pval"],
                fdr=gsdict["fdr"],
                ofname=ofname,
                pheno_pos=self.pheno_pos,
                pheno_neg=self.pheno_neg,
                color=colors,
                figsize=figsize,
            )
            g.add_axes()
            g.savefig()
            return g.fig

        elif hasattr(terms, "__len__"):  # means iterable
            terms = list(terms)
            tags = [self.results[t]["hits"] for t in terms]
            runes = [self.results[t]["RES"] for t in terms]
            t = TracePlot(
                terms=terms,
                tags=tags,
                runes=runes,
                rank_metric=ranking,
                colors=colors,
                legend_kws=legend_kws,
                ofname=ofname,
                figsize=figsize,
            )
            t.add_axes()
            t.savefig(ofname)
            return t.fig
        else:
            print("not supported input: terms")



# -*- coding: utf-8 -*-
import json
import logging
import os
import xml.etree.ElementTree as ET
from collections.abc import Iterable
from typing import Dict, List, Optional, Tuple

import requests

from gseapy.utils import DEFAULT_CACHE_PATH, unique


def gsea_cls_parser(cls: str) -> Tuple[str]:
    """Extract class(phenotype) name from .cls file.

    :param cls: the a class list instance or .cls file which is identical to GSEA input .
    :return: phenotype name and a list of class vector.
    """

    if not isinstance(cls, str) and isinstance(cls, Iterable):
        classes = list(cls)
        sample_name = unique(classes)
    elif isinstance(cls, str):
        with open(cls) as c:
            file = c.readlines()
        classes = file[2].strip().split()
        sample_name = file[1].strip("#").strip().split()

        tmp = set(sample_name) & set(classes)
        if len(tmp) < 2:  # classes and sample_name are different
            s1 = classes[0]
            for i, c in enumerate(classes):
                if c == s1:
                    classes[i] = sample_name[0]
                else:
                    classes[i] = sample_name[1]
    else:
        raise Exception("Error parsing sample name!")

    if len(sample_name) != 2:
        raise Exception("Input groups have to be 2!")

    return sample_name[0], sample_name[1], classes


def gsea_edb_parser(results_path: str) -> Dict[str, List[str]]:
    """Parse results.edb file stored under **edb** file folder.

    :param results_path: the path of results.edb file.
    :return:
        a dict contains { enrichment_term: [es, nes, pval, fdr, fwer, hit_ind]}
    """

    xtree = ET.parse(results_path)
    xroot = xtree.getroot()
    res = {}
    # dict_keys(['RANKED_LIST', 'GENESET', 'FWER', 'ES_PROFILE',
    # 'HIT_INDICES', 'ES', 'NES', 'TEMPLATE', 'RND_ES', 'RANK_SCORE_AT_ES',
    # 'NP', 'RANK_AT_ES', 'FDR'])
    for node in xroot.findall("DTG"):
        enrich_term = node.attrib.get("GENESET").split("#")[1]
        es_profile = node.attrib.get("ES_PROFILE").split(" ")
        # esnull = term.get('RND_ES').split(" ")
        hit_ind = node.attrib.get("HIT_INDICES").split(" ")
        es_profile = [float(i) for i in es_profile]
        hit_ind = [int(i) for i in hit_ind]
        # esnull = [float(i) for i in esnull ]
        es = float(node.attrib.get("ES"))
        nes = float(node.attrib.get("NES"))
        pval = float(node.attrib.get("NP"))
        fdr = float(node.attrib.get("FDR"))
        fwer = float(node.attrib.get("FWER"))
        logging.debug("Enriched Gene set is: " + enrich_term)
        res[enrich_term] = [es, nes, pval, fdr, fwer, hit_ind]
    return res


def read_gmt(path: str) -> Dict[str, List[str]]:
    """Read GMT file

    :param str path: the path to a gmt file.
    :return: a dict object
    """
    if path.lower().endswith("gmt"):
        return get_library(name=path, min_size=0, max_size=100000, gene_list=None)
    else:
        raise ValueError("Please input a gmt file")
    return


def get_library(
    name: str,
    organism: str = "Human",
    min_size: int = 0,
    max_size: int = 2000,
    save: Optional[str] = None,
    gene_list: Optional[List[str]] = None,
) -> Dict[str, List[str]]:
    """Parse gene_sets.gmt(gene set database) file or download from enrichr server.

    :param str name: the gene_sets.gmt file or an enrichr library name.
                    checkout full enrichr library name here: https://maayanlab.cloud/Enrichr/#libraries

    :param str organism: choose one from { 'Human', 'Mouse', 'Yeast', 'Fly', 'Fish', 'Worm' }.
                         This arugment has not effect if input is a `.gmt` file.

    :param min_size: Minimum allowed number of genes for each gene set. Default: 0.
    :param max_size: Maximum allowed number of genes for each gene set. Default: 2000.

    :param str save: the path to save the filtered gene set database.

    :param gene_list: if input a gene list, min and max overlapped genes between gene set and gene_list are kept.

    :return dict: Return a filtered gene set database dictionary.

    Note: **DO NOT** filter gene sets, when use :func:`replot`. Because ``GSEA`` Desktop have already
    done this for you.

    """
    genesets_dict = {}
    if name.lower().endswith(".gmt"):
        logging.info("User Defined gene sets is given.......continue..........")
        with open(name) as genesets:
            for line in genesets:
                entries = line.strip().split("\t")
                key = entries[0]
                genesets_dict[key] = entries[2:]
    else:
        # get gene sets from enrichr libary
        names = get_library_name(organism=organism)
        if name in names:
            logging.info("Downloading and generating Enrichr library gene sets...")
            genesets_dict = download_library(name, organism=organism)
        else:
            raise ValueError(
                "Sorry. The input: %s could be be found given organism: %s"
                % (name, organism)
            )

    # filtering gene_sets
    total = len(genesets_dict)
    keys = list(genesets_dict.keys())
    if gene_list is None:
        for k in keys:
            if min_size <= len(genesets_dict[k]) <= max_size:
                continue
            del genesets_dict[k]
    else:
        # given a gene_list, filter gene sets by gene_overlap numbers
        gene_dict = {g: i for i, g in enumerate(gene_list)}
        for subset in keys:
            subset_list = set(genesets_dict[subset])  # remove duplicates
            # drop genes not found in the gene_dict
            gene_overlap = [g for g in subset_list if g in gene_dict]
            tag_len = len(gene_overlap)
            if (min_size <= tag_len <= max_size) and tag_len < len(gene_list):
                # tag_len should < gene_list
                genesets_dict[subset] = gene_overlap
                continue
            del genesets_dict[subset]

    filsets_num = total - len(genesets_dict)
    if filsets_num > 0:
        logging.info(
            "%04d gene_sets have been filtered out when max_size=%s and min_size=%s"
            % (filsets_num, max_size, min_size)
        )

    if filsets_num == len(genesets_dict):
        raise Exception(
            "No gene sets passed throught filtering condition!!!, try new paramters again!\n"
            + "Note: Gene names for gseapy is case sensitive."
        )

    if save is not None:
        gmtout = open(save, "w")
        for k, v in genesets_dict.items():
            outline = "%s\t%s\t%s\n" % (k, name, "\t".join(v))
            gmtout.write(outline)
        gmtout.close()

    return genesets_dict


def get_library_name(organism: str = "Human") -> List[str]:
    """return enrichr active enrichr library name.
    see also: https://maayanlab.cloud/modEnrichr/

    :param str organism: Select one from { 'Human', 'Mouse', 'Yeast', 'Fly', 'Fish', 'Worm' }
    :return: a list of enrichr libraries from selected database

    """
    default = [
        "human",
        "mouse",
        "hs",
        "mm",
        "homo sapiens",
        "mus musculus",
        "h. sapiens",
        "m. musculus",
    ]
    _organisms = {
        "Fly": ["fly", "d. melanogaster", "drosophila melanogaster"],
        "Yeast": ["yeast", "s. cerevisiae", "saccharomyces cerevisiae"],
        "Worm": ["worm", "c. elegans", "caenorhabditis elegans", "nematode"],
        "Fish": ["fish", "d. rerio", "danio rerio", "zebrafish"],
    }
    ENRICHR_URL = "http://maayanlab.cloud"
    database = ""
    if organism.lower() in default:
        database = "Enrichr"
    else:
        for k, v in _organisms.items():
            if organism.lower() in v:
                database = k + "Enrichr"
                break

    if not database.endswith("Enrichr"):
        raise LookupError(
            """No supported database. Please input one of these:
                            ('Human', 'Mouse', 'Yeast', 'Fly', 'Fish', 'Worm') """
        )
    # make a get request to get the gmt names and meta data from Enrichr
    # old code
    # response = requests.get('http://amp.pharm.mssm.edu/Enrichr/geneSetLibrary?mode=meta')
    # gmt_data = response.json()
    # # generate list of lib names
    # libs = []
    # # get library names
    # for inst_gmt in gmt_data['libraries']:
    #     # only include active gmts
    #     if inst_gmt['isActive'] == True:
    #         libs.append(inst_gmt['libraryName'])
    lib_url = "%s/%s/datasetStatistics" % (ENRICHR_URL, database)
    response = requests.get(lib_url, verify=True)
    if not response.ok:
        raise Exception("Error getting the Enrichr libraries")
    libs_json = json.loads(response.text)
    libs = [lib["libraryName"] for lib in libs_json["statistics"]]

    return sorted(libs)


def download_library(
    name: str, organism: str = "human", filename: str = None
) -> Dict[str, List[str]]:
    """download enrichr libraries.

    :param str name: the enrichr library name. see `gseapy.get_library_name()`.
    :param str organism: Select one from { 'Human', 'Mouse', 'Yeast', 'Fly', 'Fish', 'Worm' }
    :param str filename: the file name to save if not None.
    :return dict: gene_sets of the enrichr library from selected organism


    """

    default = [
        "human",
        "mouse",
        "hs",
        "mm",
        "homo sapiens",
        "mus musculus",
        "h. sapiens",
        "m. musculus",
    ]
    _organisms = {
        "Fly": ["fly", "d. melanogaster", "drosophila melanogaster"],
        "Yeast": ["yeast", "s. cerevisiae", "saccharomyces cerevisiae"],
        "Worm": ["worm", "c. elegans", "caenorhabditis elegans", "nematode"],
        "Fish": ["fish", "d. rerio", "danio rerio", "zebrafish"],
    }
    ENRICHR_URL = "http://maayanlab.cloud"
    database = ""
    if organism.lower() in default:
        database = "Enrichr"
    else:
        for k, v in _organisms.items():
            if organism.lower() in v:
                database = k + "Enrichr"
                break

    if not database.endswith("Enrichr"):
        raise LookupError(
            """No supported database. Please input one of these:
                            ('Human', 'Mouse', 'Yeast', 'Fly', 'Fish', 'Worm') """
        )

    tmpname = "%s.%s.gmt" % (database, name)
    tempath = os.path.join(DEFAULT_CACHE_PATH, tmpname)
    if os.path.isfile(tempath):
        logging.info("Library is already downloaded in: %s, use local file" % tempath)
        genesets_dict = {}
        with open(tempath) as genesets:
            for line in genesets:
                entries = line.strip().split("\t")
                key = entries[0]
                genesets_dict[key] = entries[2:]
        return genesets_dict
    # queery string
    ENRICHR_URL = ENRICHR_URL + "/%s/geneSetLibrary" % database
    query_string = "?mode=text&libraryName=%s"
    # get
    response = requests.get(
        ENRICHR_URL + query_string % name, timeout=None, stream=True
    )
    if not response.ok:
        raise Exception(
            "Error fetching gene set library, input name is correct for the organism you've set?."
        )
    # reformat to dict
    genesets_dict = {}
    for line in response.iter_lines(chunk_size=1024, decode_unicode="utf-8"):
        line = line.strip().split("\t")
        k = line[0]
        v = map(lambda x: x.split(",")[0], line[2:])
        v = list(filter(lambda x: True if len(x) else False, v))
        genesets_dict[k] = v

    if filename is not None:
        gmtout = open(filename, "w")
        for k, v in genesets_dict.items():
            outline = "%s\t%s\t%s\n" % (k, name, "\t".join(v))
            gmtout.write(outline)
        gmtout.close()

    return genesets_dict



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# see: http://amp.pharm.mssm.edu/Enrichr/help#api for API docs
import io
import json
import logging
import os
from typing import Any, Dict, Iterable, List, Optional, Set, Tuple, Union

import numpy as np
import pandas as pd
import requests

from gseapy.biomart import Biomart
from gseapy.plot import barplot
from gseapy.stats import calc_pvalues, multiple_testing_correction
from gseapy.utils import DEFAULT_CACHE_PATH, log_init, mkdirs, retry


class EnrichrError(Exception):
    """Base exception for Enrichr-related errors."""

    pass


class EnrichrValidationError(EnrichrError):
    """Raised when input validation fails."""

    pass


class EnrichrAPIError(EnrichrError):
    """Raised when Enrichr API requests fail."""

    pass


class EnrichrNetworkError(EnrichrError):
    """Raised when network connectivity issues occur."""

    pass


class EnrichrParseError(EnrichrError):
    """Raised when parsing responses or files fails."""

    pass


class EnrichrAPI:
    """A Python client for the modEnrichr suite and Speedrichr REST APIs."""

    def __init__(self, organism: str = "human"):
        """
        Initializes the API client for a specific organism.

        Args:
            organism (str): The target organism. Options include 'human', 'mouse',
                            'fly', 'yeast', 'worm', 'fish'. Defaults to 'human'.
        """
        organism = organism.lower().strip()

        # Map common names to the correct modEnrichr instance
        self.db_map = {
            "human": "Enrichr",
            "hsapiens": "Enrichr",
            "homo sapiens": "Enrichr",
            "hs": "Enrichr",
            "h. sapiens": "Enrichr",
            "mus musculus": "Enrichr",
            "m. musculus": "Enrichr",
            "mouse": "Enrichr",
            "mm": "Enrichr",
            "enrichr": "Enrichr",
            "fly": "FlyEnrichr",
            "drosophila": "FlyEnrichr",
            "drosophila melanogaster": "FlyEnrichr",
            "d. melanogaster": "FlyEnrichr",
            "yeast": "YeastEnrichr",
            "saccharomyces": "YeastEnrichr",
            "s. cerevisiae": "YeastEnrichr",
            "saccharomyces cerevisiae": "YeastEnrichr",
            "worm": "WormEnrichr",
            "celegans": "WormEnrichr",
            "c. elegans": "WormEnrichr",
            "caenorhabditis elegans": "WormEnrichr",
            "nematode": "WormEnrichr",
            "fish": "FishEnrichr",
            "zebrafish": "FishEnrichr",
            "danio rerio": "FishEnrichr",
            "d. rerio": "FishEnrichr",
        }

        if organism not in self.db_map:
            valid_opts = ", ".join(sorted(set(self.db_map.keys())))
            raise ValueError(
                f"Invalid organism '{organism}'. Valid options are: {valid_opts}"
            )

        instance = self.db_map[organism]

        # Set the dynamic base URL
        self.base_url = f"https://maayanlab.cloud/{instance}"

        # Speedrichr URL (primarily supports standard Enrichr)
        self.speedrichr_url = "https://maayanlab.cloud/speedrichr/api"

        # Shared session with retry/pooling for all HTTP calls
        self._session = retry()

    def _ensure_list(self, item: Union[str, List[str]]) -> List[str]:
        """Utility to convert a single string to a list for uniform processing."""
        return [item] if isinstance(item, str) else item

    # ------------------------------------------------------------------------
    # Info & Utility Endpoints
    # ------------------------------------------------------------------------

    def get_libraries(self) -> List[str]:
        """Fetches a list of all available gene set library names for the current organism."""
        url = f"{self.base_url}/datasetStatistics"
        response = self._session.get(url)
        response.raise_for_status()
        data = response.json()

        # The API returns a dictionary with a 'statistics' key containing a list of library info
        return [library["libraryName"] for library in data.get("statistics", [])]

    def find_terms_by_gene(
        self, gene: str, include_json: bool = True, include_setup: bool = True
    ) -> Dict[str, Any]:
        """Finds terms that contain a given gene across Enrichr libraries."""
        url = f"{self.base_url}/genemap"
        params = {
            "gene": gene,
            "json": str(include_json).lower(),
            "setup": str(include_setup).lower(),
        }
        response = self._session.get(url, params=params)
        response.raise_for_status()
        return response.json()

    # ------------------------------------------------------------------------
    # Standard Enrichment API
    # ------------------------------------------------------------------------

    def add_list(
        self, genes: List[str], description: str = "Gene list"
    ) -> Dict[str, Any]:
        """Analyzes a gene set and returns a userListId."""
        url = f"{self.base_url}/addList"
        payload = {
            "list": (None, "\n".join(genes)),
            "description": (None, description),
        }
        response = self._session.post(url, files=payload)
        response.raise_for_status()
        return response.json()

    def view_list(self, user_list_id: int) -> Dict[str, Any]:
        """Views an added gene set using its userListId."""
        url = f"{self.base_url}/view"
        params = {"userListId": user_list_id}
        response = self._session.get(url, params=params)
        response.raise_for_status()
        return response.json()

    def get_enrichment(
        self, user_list_id: int, gene_set_library: Union[str, List[str]]
    ) -> Dict[str, Any]:
        """Gets enrichment results for one or more gene set libraries."""
        url = f"{self.base_url}/enrich"
        libraries = self._ensure_list(gene_set_library)
        all_results = {}

        for lib in libraries:
            params = {"userListId": user_list_id, "backgroundType": lib}
            response = self._session.get(url, params=params)
            response.raise_for_status()
            all_results[lib] = response.json()

        return all_results

    def get_results_dataframe(
        self, user_list_id: int, gene_set_library: Union[str, List[str]]
    ) -> pd.DataFrame:
        """Fetches enrichment analysis results and returns a combined pandas DataFrame."""
        url = f"{self.base_url}/export"
        libraries = self._ensure_list(gene_set_library)

        all_dfs = []

        for lib in libraries:
            params = {
                "userListId": user_list_id,
                "filename": "temp",
                "backgroundType": lib,
            }
            response = self._session.get(url, params=params)
            response.raise_for_status()

            df = pd.read_csv(io.StringIO(response.text), sep="\t")
            cols = df.columns.to_list()
            df["Gene_set"] = lib
            cols = ["Gene_set"] + cols  # reorder columns to put Gene_set first
            df = df[cols]
            all_dfs.append(df)
        if all_dfs:
            return pd.concat(all_dfs, ignore_index=True)
        else:
            return pd.DataFrame()

    # ------------------------------------------------------------------------
    # Background Enrichment API (Speedrichr)
    # ------------------------------------------------------------------------

    def add_list_speedrichr(
        self, genes: List[str], description: str = "Gene list with background"
    ) -> Dict[str, Any]:
        """Uploads a gene set to Speedrichr for background analysis."""
        url = f"{self.speedrichr_url}/addList"
        payload = {
            "list": (None, "\n".join(genes)),
            "description": (None, description),
        }
        response = self._session.post(url, files=payload)
        response.raise_for_status()
        return response.json()

    def add_background(self, background_genes: List[str]) -> Dict[str, Any]:
        """Uploads a background gene set to Speedrichr."""
        url = f"{self.speedrichr_url}/addbackground"
        payload = {"background": "\n".join(background_genes)}
        response = self._session.post(url, data=payload)
        response.raise_for_status()
        return response.json()

    def get_background_enrichment(
        self,
        user_list_id: int,
        background_id: str,
        gene_set_library: Union[str, List[str]],
    ) -> Dict[str, Any]:
        """Gets enrichment results calculated against a custom background for one or more libraries."""
        url = f"{self.speedrichr_url}/backgroundenrich"
        libraries = self._ensure_list(gene_set_library)
        all_results = []

        for lib in libraries:
            payload = {
                "userListId": user_list_id,
                "backgroundid": background_id,
                "backgroundType": lib,
            }
            response = self._session.post(url, data=payload)
            response.raise_for_status()
            df = self._parse_json_response(response.json(), lib)
            cols = df.columns.to_list()
            df["Gene_set"] = lib
            cols = ["Gene_set"] + cols  # reorder columns to put Gene_set first
            df = df[cols]
            all_results.append(df)
        all_results = (
            pd.concat(all_results, ignore_index=True) if all_results else pd.DataFrame()
        )
        return all_results

    def _parse_json_response(self, data: Dict, gene_set_key: str) -> pd.DataFrame:
        """Parse Enrichr JSON response into DataFrame."""
        colnames = [
            "Rank",
            "Term",
            "P-value",
            "Odds Ratio",
            "Combined Score",
            "Genes",
            "Adjusted P-value",
            "Old P-value",
            "Old adjusted P-value",
        ]
        res = pd.DataFrame(data[gene_set_key], columns=colnames)
        res["Genes"] = res["Genes"].apply(";".join)
        colord = [
            "Term",
            "P-value",
            "Adjusted P-value",
            "Old P-value",
            "Old adjusted P-value",
            "Odds Ratio",
            "Combined Score",
            "Genes",
        ]
        return res.loc[:, colord]


class Enrichr(EnrichrAPI):
    """Enrichr API"""

    def __init__(
        self,
        gene_list: Iterable[str],
        gene_sets: Union[List[str], str, Dict[str, str]],
        organism: str = "human",
        outdir: Optional[str] = "Enrichr",
        background: Union[List[str], int, str] = None,
        cutoff: float = 0.05,
        format: str = "pdf",
        figsize: Tuple[float, float] = (6.5, 6),
        top_term: int = 10,
        no_plot: bool = False,
        verbose: bool = False,
    ):
        self.gene_list = gene_list
        self.gene_sets = gene_sets
        self.descriptions = ""
        self.outdir = outdir
        self.cutoff = cutoff
        self.format = format
        self.figsize = figsize
        self.__top_term = int(top_term)
        self.__no_plot = no_plot
        self.verbose = bool(verbose)
        self.module = "enrichr"
        self.res2d: Optional[pd.DataFrame] = None
        self.background = background
        self._bg: Union[Set[str], int, None] = None
        self.organism = organism
        self._gene_isupper = True
        self._gene_toupper = False
        self._gls: Union[Set[int], List[str]] = []
        self._isezid = False
        self._gs: str = ""
        self._gs_name: List[str] = []
        super().__init__(organism)
        # init logger
        self.prepare_outdir()

    def __enter__(self):
        """Context manager entry."""
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Context manager exit."""
        self.close()
        return False

    def close(self):
        """Explicitly close logger handlers."""
        if hasattr(self, "_logger"):
            handlers = self._logger.handlers[:]
            for handler in handlers:
                handler.close()
                self._logger.removeHandler(handler)

    def prepare_outdir(self):
        """create temp directory."""
        self._outdir = self.outdir

        logfile = None
        if isinstance(self.outdir, str):
            mkdirs(self.outdir)
            logfile = os.path.join(self.outdir, f"gseapy.{self.module}.{id(self)}.log")
        self._logger = log_init(
            name=f"{self.module}{id(self)}",
            log_level=logging.INFO if self.verbose else logging.WARNING,
            filename=logfile,
        )

    def set_organism(self) -> None:
        """Initialize EnrichrAPI base with the selected organism, setting base_url."""
        if self.organism not in self.db_map:
            valid_opts = ", ".join(sorted(set(self.db_map.keys())))
            raise ValueError(
                f"Invalid organism '{self.organism}'. Valid options are: {valid_opts}"
            )
        instance = self.db_map[self.organism]
        # Set the dynamic base URL
        self.base_url = f"https://maayanlab.cloud/{instance}"

    def _parse_gmt(self, g: str) -> Dict[str, List[str]]:
        """Parse GMT file efficiently with single split per line."""
        g_dict = {}
        with open(g) as genesets:
            for line in genesets:
                parts = line.strip().split("\t")
                if len(parts) > 2:
                    g_dict[parts[0]] = parts[2:]
        return g_dict

    def _gs2dict(self, gene_sets: List[str]) -> List[Dict[str, List[str]]]:
        """helper function, only convert gmt to dict and keep strings"""
        gss = []
        self._gs_name = []
        for i, g in enumerate(gene_sets):
            # only convert gmt to dict. local mode
            if isinstance(g, str) and g.lower().endswith(".gmt"):
                if os.path.exists(g):
                    self._logger.info(f"User defined gene sets is given: {g}")
                    gss.append(self._parse_gmt(g))
                    self._gs_name.append(os.path.basename(g))
                else:
                    self._logger.warning(
                        f"User defined gene sets is not found: {g}, skip."
                    )
            else:
                gss.append(g)
                _name = g
                if isinstance(g, dict):
                    _name = f"gs_ind_{i}"
                    self._logger.info(f"Input dict object named with {_name}")
                self._gs_name.append(_name)
        return gss

    def parse_genesets(self, gene_sets=None) -> List[Union[Dict[str, List[str]], str]]:
        """parse gene_sets input file type"""
        if gene_sets is None:
            gene_sets = self.gene_sets

        gss = []
        if isinstance(gene_sets, list):
            gss = self._gs2dict(gene_sets)
        elif isinstance(self.gene_sets, str):
            gss = [g.strip() for g in gene_sets.strip().split(",")]
            gss = self._gs2dict(gss)
        elif isinstance(gene_sets, dict):
            gss = self._gs2dict([gene_sets.copy()])
        else:
            raise EnrichrValidationError(
                "Error parsing enrichr libraries, please provided corrected one"
            )

        # now, gss[LIST] contains dict or strings.
        if len(gss) < 1:
            raise EnrichrValidationError(
                "No GeneSets are valid !!! Check your gene_sets input."
            )
        gss_exist = []
        gss_name = []
        enrichr_library = []
        # if all local gmts (local mode), skip connect to enrichr server
        if not all([isinstance(g, dict) for g in gss]):
            enrichr_library = self.get_libraries()

        # check enrichr libraries are valid
        for n, g in zip(self._gs_name, gss):
            if isinstance(g, dict):
                gss_exist.append(g)
                gss_name.append(n)
                continue
            if isinstance(g, str):
                if g in enrichr_library:
                    gss_exist.append(g)
                    gss_name.append(n)
                else:
                    self._logger.warning(f"Input library not found: {g}. Skip")

        self._gs_name = gss_name  # update names
        if len(gss_exist) < 1:
            raise EnrichrValidationError(
                "No GeneSets are valid !!! Check your gene_sets input."
            )
        return gss_exist

    def parse_genelists(self) -> str:
        """Parse gene list with single-pass processing."""
        if isinstance(self.gene_list, list):
            genes = self.gene_list
        elif isinstance(self.gene_list, pd.DataFrame):
            # input type is bed file
            if self.gene_list.shape[1] >= 3:
                genes = (
                    self.gene_list.iloc[:, :3]
                    .apply(lambda x: "\t".join([str(i) for i in x]), axis=1)
                    .to_list()
                )
            # input type with weight values
            elif self.gene_list.shape[1] == 2:
                genes = self.gene_list.apply(
                    lambda x: ",".join([str(i) for i in x]), axis=1
                ).to_list()
            else:
                genes = self.gene_list.squeeze().to_list()
        elif isinstance(self.gene_list, pd.Series):
            genes = self.gene_list.squeeze().to_list()
        else:
            # get gene lists or bed file, or gene list with weighted values.
            genes = []
            with open(self.gene_list) as f:
                for gene in f:
                    genes.append(gene.strip())

        if not genes:
            raise EnrichrValidationError("Gene list cannot be empty")

        # Single-pass processing: strip whitespace and check if entrez IDs
        cleaned_genes = []
        all_entrez = True
        for g in genes:
            cleaned = g.strip()
            cleaned_genes.append(cleaned)
            if all_entrez and not self._is_entrez_id(cleaned):
                all_entrez = False

        self._isezid = all_entrez
        if self._isezid:
            self._gls = set(map(int, cleaned_genes))
        else:
            self._gls = cleaned_genes

        return "\n".join(cleaned_genes)

    def check_uppercase(self, gene_list: List[str]) -> bool:
        """
        Check whether a list of gene names are mostly in uppercase.

        Parameters
        ----------
        gene_list : list
            A list of gene names

        Returns
        -------
        bool
            Whether the list of gene names are mostly in uppercase
        """
        if all(self._is_entrez_id(s) for s in gene_list):
            return False
        is_upper = [str(s).isupper() for s in gene_list]
        return sum(is_upper) / len(is_upper) >= 0.9

    def _is_entrez_id(self, idx: Union[int, str]) -> bool:
        """Check if the input is a valid Entrez ID (integer)."""
        try:
            int(idx)
            return True
        except (ValueError, TypeError):
            return False

    def send_genes(self, payload, url) -> Dict[str, Union[int, str]]:
        """Send gene list to enrichr server."""
        try:
            response = self._session.post(url, files=payload, verify=True)
            if not response.ok:
                self._logger.debug(f"URL: {url}")
                self._logger.debug(f"Payload: {payload}")
                raise EnrichrAPIError(
                    f"Error sending gene list, status code: {response.status_code}"
                )
            job_id = json.loads(response.text)
            # response.text format:
            # {"userListId": 667152768, "shortId": "27c3f180"}
            return job_id
        except requests.exceptions.RequestException as e:
            raise EnrichrNetworkError(f"Network error sending gene list: {e}")

    def check_genes(self, gene_list: List[str], usr_list_id: str) -> None:
        """Compare the genes sent and received to get successfully recognized genes."""
        try:
            returned = self.view_list(usr_list_id)
            returnedL = returned["genes"]
            returnedN = sum(1 for gene in gene_list if gene in returnedL)
            self._logger.info(f"{returnedN} genes successfully recognized by Enrichr")
        except requests.exceptions.RequestException as e:
            raise EnrichrNetworkError(f"Network error checking genes: {e}")

    def get_results_with_background(
        self, gene_list: str, background: List[str], gene_set_libraries: List[str]
    ) -> Tuple[str, pd.DataFrame]:
        """Get enrichment results with custom background."""
        payload = {"list": (None, gene_list), "description": (None, self.descriptions)}
        job_id = self.send_genes(payload, f"{self.speedrichr_url}/addList")

        bg_id = self.add_background(background)

        res = self.get_background_enrichment(
            job_id["userListId"], bg_id["backgroundid"], gene_set_libraries
        )
        return (job_id["shortId"], res)

    def get_results(
        self, gene_list: str, gene_set_libraries: List[str]
    ) -> Tuple[str, pd.DataFrame]:
        """Get enrichment results from Enrichr API."""
        payload = {"list": (None, gene_list), "description": (None, self.descriptions)}
        job_id = self.send_genes(payload, f"{self.base_url}/addList")

        res = self.get_results_dataframe(job_id["userListId"], gene_set_libraries)
        return (job_id["shortId"], res)

    def get_background(self) -> Set[str]:
        """Get background genes from file or BioMart."""
        # input is a file
        if os.path.isfile(self.background):
            with open(self.background) as b:
                bg = [g.strip() for g in b]
            return set(bg)

        # package included data
        db_file = os.path.join(
            DEFAULT_CACHE_PATH, f"{self.background}.background.genes.txt"
        )
        if os.path.exists(db_file):
            df = pd.read_csv(db_file, sep="\t")
        else:
            # background is a biomart database name
            self._logger.warning(
                f"Downloading {self.background} for the first time. It might take a couple of minutes."
            )
            bm = Biomart()
            df = bm.query(
                dataset=self.background,
                attributes=["ensembl_gene_id", "external_gene_name", "entrezgene_id"],
                filename=db_file,
            )
        self._logger.info(
            f"Using all annotated genes with GO_ID as background: {self.background}"
        )
        # input id type: entrez or gene_name
        if self._isezid:
            df.dropna(subset=["entrezgene_id"], inplace=True)
            bg = df["entrezgene_id"].astype(int)
        else:
            df.dropna(subset=["external_gene_name"], inplace=True)
            bg = df["external_gene_name"]

        return set(bg)

    def filter_gmt(
        self, gmt: Dict[str, List[str]], background: Set[str]
    ) -> Dict[str, List[str]]:
        """Filter GMT to only include genes that exist in background.

        This substantially affects the significance of the hypergeometric test.

        Parameters
        ----------
        gmt : dict
            A dict of gene sets
        background : set
            A set of custom background genes

        Returns
        -------
        dict
            Filtered gene sets
        """
        gmt_filtered = {}
        for term, genes in gmt.items():
            filtered_genes = [g for g in genes if g in background]
            if filtered_genes:
                gmt_filtered[term] = filtered_genes
        return gmt_filtered

    def parse_background(
        self, gmt: Optional[Dict[str, List[str]]] = None
    ) -> Union[Set[str], int]:
        """Parse and set background genes."""
        self._bg = set()
        if self.background is None:
            # use all genes in the dict input as background if background is None
            if gmt:
                bg = set()
                for _, genes in gmt.items():
                    bg = bg.union(set(genes))
                self._logger.info(
                    f"  Background is not set! Use all {len(bg)} genes in {self._gs}."
                )
                self._bg = bg
        elif np.isscalar(self.background):
            if isinstance(self.background, int) or self.background.isdigit():
                self._bg = int(self.background)
                self._logger.info(f"  Background input is a number: {self._bg}")
            elif isinstance(self.background, str):
                self._bg = self.get_background()
                self._logger.info(f"  Background: found {len(self._bg)} genes")
            else:
                raise EnrichrValidationError("Unsupported background data type")
        else:
            # handle array object: nd.array, list, tuple, set, Series
            try:
                _ = iter(self.background)
                self._bg = set(self.background)
            except TypeError:
                self._logger.error("  Unsupported background data type")
                raise EnrichrValidationError("Unsupported background data type")

        return self._bg

    def enrich_local(self, gmt: Dict[str, List[str]]) -> Optional[pd.DataFrame]:
        """Perform local enrichment analysis using hypergeometric test.

        p-value: computed using the Fisher exact test (Hypergeometric test)
        z-score: Odds Ratio
        combined score: -log(p) * z

        See: http://amp.pharm.mssm.edu/Enrichr/help#background&q=4

        Columns: Term, Overlap, P-value, Odds Ratio, Combined Score, Adjusted_P-value, Genes
        """
        # Check gene case consistency
        top10 = min(len(gmt), 10)
        top10_keys = list(gmt.keys())[:top10]
        _gls = self._gls
        ups = [self.check_uppercase(gmt[key]) for key in top10_keys]
        _gene_toupper = False

        if all(ups) and not self._gene_isupper:
            _gls = [str(x).upper() for x in self._gls]
            _gene_toupper = True
            self._logger.info(
                "  Genes in GMT file are all in upper case, convert query to upper case."
            )

        bg = self.parse_background(gmt)
        if isinstance(bg, set) and not self._gene_isupper and _gene_toupper:
            bg = {str(s).upper() for s in bg}

        # Statistical testing
        hgtest = list(calc_pvalues(query=_gls, gene_sets=gmt, background=bg))
        if len(hgtest) > 0:
            terms, pvals, oddr, olsz, gsetsz, genes = hgtest
            fdrs, _ = multiple_testing_correction(
                ps=pvals, alpha=self.cutoff, method="benjamini-hochberg"
            )
            # Build result DataFrame (dict maintains insertion order in Python 3.7+)
            res = pd.DataFrame(
                {
                    "Gene_set": self._gs,
                    "Term": terms,
                    "Overlap": [f"{h}/{g}" for h, g in zip(olsz, gsetsz)],
                    "P-value": pvals,
                    "Adjusted P-value": fdrs,
                    "Odds Ratio": oddr,
                    "Combined Score": -1 * np.log(pvals) * oddr,
                    "Genes": [";".join(map(str, g)) for g in genes],
                }
            )
            return res
        return None

    def enrich_online(
        self, genes_list: str, geneset_libraries: List[str]
    ) -> Optional[pd.DataFrame]:
        """Perform online enrichment analysis using Enrichr API."""

        # Online mode - API enrichment
        self._logger.debug(f"Enrichr service using library: {geneset_libraries}")
        bg = self.parse_background()
        # Whether user input background
        if isinstance(bg, set) and len(bg) > 0:
            _shortID, res = self.get_results_with_background(
                genes_list, list(bg), geneset_libraries
            )
        else:
            _shortID, res = self.get_results(genes_list, geneset_libraries)
        return res

    def _save_results(self, res: pd.DataFrame, name: str) -> None:
        """Save enrichment results to file and generate plots."""
        if self._outdir is None:
            return

        outfile = f"{self.outdir}/{name}.{self.organism}.{self.module}.reports.txt"
        res.to_csv(
            outfile, index=False, encoding="utf-8", float_format="%.6e", sep="\t"
        )
        self._logger.info(f"Save enrichment results for {name}")

        # Generate plots
        if not self.__no_plot:
            _ax = barplot(
                df=res,
                cutoff=self.cutoff,
                figsize=self.figsize,
                top_term=self.__top_term,
                color="salmon",
                title=name,
                ofname=outfile.replace("txt", self.format),
            )
            self._logger.debug("Generate figures")

    def _validate_inputs(self) -> Tuple[str, List[Union[Dict[str, List[str]], str]]]:
        """Validate and parse inputs."""
        # Parse gene list
        genes_list = self.parse_genelists()

        if not self._isezid:
            self._gene_isupper = self.check_uppercase(self._gls)

        # Parse gene sets
        gss = self.parse_genesets()
        if len(gss) < 1:
            self._logger.error(f"None of your input gene set matched! {self.gene_sets}")
            self._logger.error(
                f"Hint: Current organism = {self.organism}, is this correct?\n"
                "Hint: use get_library_name() to view full list of supported names."
            )
            raise LookupError(
                "Not validated Enrichr library! Please provide correct organism and library name!"
            )

        return genes_list, gss

    def run(self) -> None:
        """Run enrichr for one sample gene list against multiple libraries."""
        # Validate inputs
        genes_list, gss = self._validate_inputs()

        self.results = []

        online_genesets = [g for g in gss if isinstance(g, str)]
        # run online enrichment for all string libraries together to minimize API calls
        if online_genesets:
            self._logger.info(
                f"Online enrichment analysis with libraries: {', '.join(online_genesets)}"
            )
            res_online = self.enrich_online(genes_list, online_genesets)
            if res_online is not None:
                self.results.append(res_online)

        ## Process local gene sets (dicts) separately to avoid unnecessary API calls
        for name, geneset in zip(self._gs_name, gss):
            if isinstance(geneset, dict):
                # Local mode - offline enrichment
                self._logger.info(f"Off-line enrichment analysis with library: {name}")
                if self._isezid:
                    geneset = {k: list(map(int, v)) for k, v in geneset.items()}
                self._gs = (
                    name  # assign the name of local gene set for background parsing
                )
                res_local = self.enrich_local(geneset)
                if res_local is not None:
                    self.results.append(res_local)

        if len(self.results) == 0:
            self._logger.error("No hits returned for all input gene sets!")
            return

        self.results = pd.concat(self.results, ignore_index=True)
        self.res2d = self.results
        self._save_results(self.results, name="Enrichr")
        self._logger.info("Done.")
